/* $Id: mpi.c,v 1.1.2.7 1997/03/19 09:11:39 ecd Exp $
 * mpi.c: Probing for the Tadpole SPARCbook 1 MPI.
 *
 * Copyright (C) 1996 Eddie C. Dost (ecd@skynet.be)
 */

#define __KERNEL_SYSCALLS__

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/unistd.h>

#include <asm/oplib.h>
#include <asm/io.h>
#include <asm/mpi.h>

extern void tadkbd_inchar(unsigned char ch, struct pt_regs *regs);


struct mpi_host_cmd {
	unsigned char cmd;
	int	      nin, nout;
};

static struct mpi_host_cmd mpi_host_cmd_table[MPI_MAX_HOST_CMD+1] = {
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ MPI_HOST_CMD_SER_NO, 0, 3 },
	{ MPI_HOST_CMD_HW_REV, 0, 2 },
	{ MPI_HOST_CMD_FW_REV, 0, 2 },
	{ MPI_HOST_CMD_MAXTEMP, 0, 1 },
	{ MPI_HOST_CMD_MINTEMP, 0, 1 },
	{ MPI_HOST_CMD_TEMP, 0, 1 },
	{ MPI_HOST_CMD_POW_COUNT, 0, 2 },
	{ MPI_HOST_CMD_POW_SEC, 0, 4 },
	{ MPI_HOST_CMD_STATUS, 0, 1 },
	{ MPI_HOST_CMD_ADC, 1, 2 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ MPI_HOST_CMD_LED_OFF, 0, 0 },
	{ MPI_HOST_CMD_LED_ON, 0, 0 },
	{ MPI_HOST_CMD_BEEP, 1, 0 },
	{ MPI_HOST_CMD_INC_BRIGHT, 1, 0 },
	{ MPI_HOST_CMD_DEC_BRIGHT, 1, 0 },
	{ MPI_HOST_CMD_INC_CONT, 1, 0 },
	{ MPI_HOST_CMD_DEC_CONT, 1, 0 },
	{ MPI_HOST_CMD_RD_EEPROM, 1, 2 },
	{ MPI_HOST_CMD_WR_EEPROM, 3, 0 },
	{ MPI_HOST_CMD_KBD_REP, 2, 0 },
	{ MPI_HOST_CMD_MS_SCAN, 1, 0 },
	{ MPI_HOST_CMD_WR_PORT1, 1, 0 },
	{ MPI_HOST_CMD_RD_PORT1, 1, 1 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ MPI_HOST_CMD_ACK_INTR, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ 0xff, 0, 0 }, { 0xff, 0, 0 }, { 0xff, 0, 0 },
	{ MPI_HOST_CMD_WD_ENABLE, 0, 0 },
	{ MPI_HOST_CMD_WD_DISABLE, 0, 0 },
	{ MPI_HOST_CMD_WD_CMD, 0, 0 }
};


struct mpi_regs *mpi_regs;

struct mpi_cmd_struct {
	int			 busy;
	int			 cmd;
	int			 status;
	int			 count;
	int			 pending;
	unsigned char		*buffer;
	struct wait_queue	*wait;
	struct timer_list 	 timer;
};


static void mpi_timeout(unsigned long id)
{
	struct mpi_cmd_struct *cmd = (struct mpi_cmd_struct *)id;

	printk("MPI: command %02x timeout\n", cmd->cmd);
	cmd->pending = 0;
	wake_up(&cmd->wait);
}


static struct mpi_cmd_struct mpi_command = {
	0,			/* busy		*/
	0,			/* cmd		*/
	0,			/* status	*/
	0,			/* count	*/
	0,			/* pending	*/
	NULL,			/* buffer	*/
	NULL,			/* wait		*/
	{ NULL, NULL, 0, (unsigned long)&mpi_command, mpi_timeout }
};


static void
mpi_write_char(unsigned char ch)
{
	int i;

	while ((mpi_regs->tx_status & 0x01) == 0x01)
		/* wait */;
	for (i = 0; i < 25 * 200 + 1000; i++)
		/* delay */;
	mpi_regs->tx_data = ch;
}


int mpi_host_command(unsigned char cmd, unsigned char *in, int inlen,
		     unsigned char *out, int outlen)
{
	struct wait_queue wait = { current, NULL };
	struct mpi_host_cmd *h;
	int i;

	if (cmd > MPI_MAX_HOST_CMD) {
		printk("MPI: Unknown host command %02x\n", cmd);
		return -1;
	}

	h = &mpi_host_cmd_table[cmd];
	if (h->cmd == 0xff) {
		printk("MPI: Unknown host command %02x\n", cmd);
		return -1;
	}

	if (h->nin != inlen) {
		printk("MPI: wrong number of inputs %d, expected %d\n",
		       inlen, h->nin);
		return -1;
	}

	if (h->nout > outlen) {
		printk("MPI: short number of outputs %d, need %d\n",
		       outlen, h->nout);
		return -1;
	}

	if (set_bit(0, &mpi_command.busy) != 0) {
		printk("MPI: busy\n");
		return -1;
	}

	mpi_command.cmd = cmd;
	mpi_command.status = -1;
	mpi_command.count = h->nout;
	mpi_command.buffer = out;
	add_wait_queue(&mpi_command.wait, &wait);

	del_timer(&mpi_command.timer);

	mpi_write_char(h->cmd);
	for (i = 0; i < h->nin; i++)
		mpi_write_char(in[i]);

	mpi_command.timer.expires = jiffies + HZ;
	add_timer(&mpi_command.timer);

	current->state = TASK_INTERRUPTIBLE;
	schedule();

	remove_wait_queue(&mpi_command.wait, &wait);

	mpi_command.busy = 0;
	return mpi_command.status;
}


static void mpi_interrupt(int irq, void *id, struct pt_regs *regs)
{
	static int mouse_input_pending = 0;
	unsigned char val;

	val = mpi_regs->kbd_data;
	mpi_regs->int_ack = (1 << irq);

	if (mpi_command.pending) {
		*mpi_command.buffer++ = val;
		mpi_command.pending--;
		if (!mpi_command.pending) {
			mpi_command.status = 0;
			del_timer(&mpi_command.timer);
			wake_up(&mpi_command.wait);
		}
		return;
	}

	if (mouse_input_pending) {
		mouse_input_pending--;
		return;
	}

	if (val == 0xfb) {
		/*
		 * Well, we get these at boot time.
		 */
		return;
	}

	if (val == 0xfd) {
		/*
		 * Input from the force sensitive resistors, send
		 * next 3 chars to the mouse driver.
		 */
		mouse_input_pending = 3;
	}

	if (val == 0xfe) {
		/*
		 * MPI host command has been answered.
		 */
		if (!mpi_command.count) {
			mpi_command.status = 0;
			del_timer(&mpi_command.timer);
			wake_up(&mpi_command.wait);
			return;
		}
		mpi_command.pending = mpi_command.count;
		return;
	}

	/*
	 * Anything else we through at the keyboard driver...
	 */
	tadkbd_inchar(val, regs);
}


static struct timer_list sysmon_timer;
static struct wait_queue *sysmond_wait = NULL;


static void
sysmon_tick(unsigned long unused)
{
	del_timer(&sysmon_timer);
	sysmon_timer.expires = jiffies + 60 * HZ;
	add_timer(&sysmon_timer);
	wake_up(&sysmond_wait);
}

static void
init_sysmon_timer(void)
{
	sysmon_timer.next = sysmon_timer.prev = NULL;
	sysmon_timer.expires = jiffies + 60 * HZ;
	sysmon_timer.data = 0;
	sysmon_timer.function = sysmon_tick;
	add_timer(&sysmon_timer);
}
	

static int
sysmond(void *unused)
{
	char *revision = "$Revision: 1.1.2.7 $", *s, *e;
	unsigned char chan, out[2];
	unsigned short temp, bat;
	int i;

	current->session = 1;
	current->pgrp = 1;
	sprintf(current->comm, "sysmond");
	current->blocked = ~0UL;

#ifdef __SMP__
	lock_kernel();
	syscall_count++;
#endif

	init_sysmon_timer();

	if ((s = strchr(revision, ':')) &&
	    (e = strchr(s, '$')))
		s += 2, e--, i = e - s;
	else
		s = revision, i = -1;
	printk("Started sysmond v%.*s\n", i, s);

	while (1) {
		current->signal = 0;
		interruptible_sleep_on(&sysmond_wait);

		chan = 0x02;
		mpi_host_command(MPI_HOST_CMD_ADC, &chan, 1, out, 2);
		temp = (out[0] << 2) | (out[1] & 0x03);

		/*
		 * If temperature is getting high, half the MBus clock.
		 * If this all doesn't help, savely shut down the system.
		 */

		chan = 0x03;
		mpi_host_command(MPI_HOST_CMD_ADC, &chan, 1, out, 2);
		bat = (out[0] << 2) | (out[1] & 0x03);

		/*
		 * If batteries start getting low, start flashing the
		 * batteries LED. If power drops below some limit, savely
		 * shut down the system.
		 */
	}
}


__initfunc(void mpi_probe(void))
{
	struct linux_prom_registers regs[1];
	int node;

	mpi_regs = 0;

	if (!((node = prom_getchild(prom_root_node)) &&
	      (node = prom_searchsiblings(node, "obio")) &&
	      (node = prom_getchild(node)) &&
	      (node = prom_searchsiblings(node, "mpi"))))
		return;

	prom_getproperty(node, "reg", (char *)regs, sizeof(regs));
	prom_apply_obio_ranges(regs, 0x1);

	/* Map the registers read/write */
	mpi_regs = (struct mpi_regs *)sparc_alloc_io(regs[0].phys_addr, 0,
						     regs[0].reg_size,
						     "MPI",
						     regs[0].which_io, 0x0);

	printk ("MPI@%x,%lx at %08lx\n", regs[0].which_io,
		(unsigned long)regs[0].phys_addr, (unsigned long)mpi_regs);
}


__initfunc(void mpi_init(void))
{
	struct linux_prom_irqs irq[1];
	int len, node;

	if (!((node = prom_getchild(prom_root_node)) &&
	      (node = prom_searchsiblings(node, "obio")) &&
	      (node = prom_getchild(node)) &&
	      (node = prom_searchsiblings(node, "mpi"))))
		panic ("MPI: probed, but node is gone (???)\n");

	len = prom_getproperty(node, "intr", (char *)irq, sizeof(irq));
	if (len %  sizeof(struct linux_prom_irqs)) {
		prom_printf("WHOOPS: proplen for %s was %d, "
			    "need multiple of %d\n", "intr",
			    len, sizeof(struct linux_prom_irqs));
		panic("MPI: intr property");
	}

	if (request_irq(irq[0].pri, &mpi_interrupt,
			SA_SHIRQ, "MPI", (void *)node)) {
		panic("MPI: unable to attach interrupt");
	}

	init_waitqueue(&mpi_command.wait);
	kernel_thread(sysmond, NULL, 0);

	printk ("MPI: irq %d\n", irq[0].pri);
}
