/* $Id: tadpole_irq.c,v 1.1.2.4 1997/03/18 07:59:47 ecd Exp $
 * tadpole_irq.c: Interrupt handler configuration on SPARCbook 1.
 *
 *  Copyright (C) 1996 Eddie C. Dost (ecd@skynet.be)
 */

#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/linkage.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>
#include <linux/init.h>

#include <asm/ptrace.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/psr.h>
#include <asm/vaddrs.h>
#include <asm/timer.h>
#include <asm/openprom.h>
#include <asm/oplib.h>
#include <asm/traps.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mpi.h>

static void tadpole_disable_irq(unsigned int irq_nr)
{
	unsigned long flags;
    
	save_and_cli(flags);
	irq_nr &= 0x0f;
	mpi_regs->int_enable &= ~(1 << irq_nr);
	restore_flags(flags);
}

static void tadpole_enable_irq(unsigned int irq_nr)
{
	unsigned long flags;
    
	save_and_cli(flags);
	irq_nr &= 0x0f;
	mpi_regs->int_enable |= (1 << irq_nr);
	restore_flags(flags);
}

#define TIMER_IRQ  	13    /* Also at level 13, but we ignore that one. */
#define PROFILE_IRQ	12    /* Level13 ticker ...                        */

static void tadpole_clear_clock_irq(void)
{
	mpi_regs->int_ack = (1 << TIMER_IRQ);
}

static void tadpole_clear_profile_irq(void)
{
	mpi_regs->int_ack = (1 << PROFILE_IRQ);
}

static void tadpole_load_profile_irq(unsigned int limit)
{
 	/* The timer is in microseconds, not shure if this is correct. */
	mpi_regs->ct1_time = -limit;
	mpi_regs->ct1_cntl = 0xe;
}

__initfunc(static void tadpole_init_timers(void (*counter_fn)(int, void *, struct pt_regs *)))
{
	int irq;

	/* Have the level 12 timer tick at 100HZ.  We don't touch the
	 * level 13 timer limit.
	 */
	mpi_regs->ct0_time = -10000;
	mpi_regs->ct0_cntl = 0x7;

	master_l10_counter = (unsigned int *)&mpi_regs->ct0_value;
	master_l10_limit = (unsigned int *)&mpi_regs->ct0_time;

	irq = request_irq(TIMER_IRQ,
			  counter_fn,
			  (SA_INTERRUPT | SA_STATIC_ALLOC),
			  "timer", NULL);
	if (irq) {
		prom_printf("time_init: unable to attach IRQ%d\n",TIMER_IRQ);
		prom_halt();
	}
    
	claim_ticker14(NULL, PROFILE_IRQ, 0);
}

#ifdef __SMP__
static void tadpole_nop(void) {}
#endif

__initfunc(void tadpole_init_IRQ(void))
{
	if (!mpi_regs)
		panic("Tadpole IRQ: MPI not mapped\n");

	mpi_regs->int_polar = 0x00000120;
	mpi_regs->int_enable = 0x00000000;

	enable_irq = tadpole_enable_irq;
	disable_irq = tadpole_disable_irq;
	clear_clock_irq = tadpole_clear_clock_irq;
	clear_profile_irq = tadpole_clear_profile_irq;
	load_profile_irq = tadpole_load_profile_irq;
	init_timers = tadpole_init_timers;
#ifdef __SMP__
	set_cpu_int = (void (*) (int, int))tadpole_nop;
	clear_cpu_int = (void (*) (int, int))tadpole_nop;
	set_irq_udt = (void (*) (int))tadpole_nop;
#endif
	sti();
}
