/*
 * Definitions for IEEE Double Precision
 */
#ifndef DOUBLE_H
#define DOUBLE_H

/* ensure that the FP headers get included in the right order */
#include "soft-fp.h"

/* this one has to stay a macro, so the scope of the declaration is correct */
#define FP_DECL_D(X)		_FP_DECL(__FP_D_TYPE,X)

/* All these functions are just wrappers. We should call the generic
 * routines passing the number of words in a double (here __FP_D_SIZE) and casting
 * our _FP_D_TYPE parameters to _FP_VAL. 
 */
INLINE void FP_UNPACK_RAW_D(__FP_D_TYPE X, void *val)
{
   _FP_UNPACK_RAW(__FP_D_SIZE,'d',(_FP_VAL)X,val);
}

INLINE void FP_PACK_RAW_D(void *val, __FP_D_TYPE X)
{
   _FP_PACK_RAW(__FP_D_SIZE,'d',val,(_FP_VAL)X);
}

INLINE void FP_UNPACK_D(__FP_D_TYPE X, void *val)
{
   _FP_UNPACK_RAW(__FP_D_SIZE,'d',(_FP_VAL)X,val);
   _FP_UNPACK_CANONICAL('d',__FP_D_SIZE,(_FP_VAL)X);
}

INLINE void FP_PACK_D(void *val, __FP_D_TYPE X)
{
   _FP_PACK_CANONICAL('d',__FP_D_SIZE,(_FP_VAL)X);
   _FP_PACK_RAW(__FP_D_SIZE,'d',val,(_FP_VAL)X);
}

INLINE void FP_NEG_D(__FP_D_TYPE R, __FP_D_TYPE X)
{
   _FP_NEG('d',__FP_D_SIZE,(_FP_VAL)R,(_FP_VAL)X);
}

INLINE void FP_ADD_D(__FP_D_TYPE R, __FP_D_TYPE X, __FP_D_TYPE Y)
{
   _FP_ADD('d',__FP_D_SIZE,(_FP_VAL)R,(_FP_VAL)X,(_FP_VAL)Y);
}

INLINE void FP_SUB_D(__FP_D_TYPE R, __FP_D_TYPE X, __FP_D_TYPE Y)
{
   /* _FP_SUB doesn't exist! 
    *   _FP_SUB('d',__FP_D_SIZE,(_FP_VAL)R,(_FP_VAL)X,(_FP_VAL)Y);
    */
}

INLINE void FP_MUL_D(__FP_D_TYPE R, __FP_D_TYPE X, __FP_D_TYPE Y)
{
   _FP_MUL('d',__FP_D_SIZE,(_FP_VAL)R,(_FP_VAL)X,(_FP_VAL)Y);
}

INLINE void FP_DIV_D(__FP_D_TYPE R, __FP_D_TYPE X, __FP_D_TYPE Y)
{
   _FP_DIV('d',__FP_D_SIZE,(_FP_VAL)R,(_FP_VAL)X,(_FP_VAL)Y);
}

INLINE void FP_SQRT_D(__FP_D_TYPE R, __FP_D_TYPE X)
{
   _FP_SQRT('d',__FP_D_SIZE,(_FP_VAL)R,(_FP_VAL)X);
}

INLINE _FP_I_TYPE FP_CMP_D(__FP_D_TYPE X, __FP_D_TYPE Y, _FP_I_TYPE un)
{
   return _FP_CMP('d',__FP_D_SIZE,(_FP_VAL)X,(_FP_VAL)Y,un);
}

INLINE _FP_I_TYPE FP_CMP_EQ_D(__FP_D_TYPE X, __FP_D_TYPE Y, _FP_I_TYPE un)
{
   return _FP_CMP_EQ('d',__FP_D_SIZE,(_FP_VAL)X,(_FP_VAL)Y,un);
}

INLINE _FP_CONV_I_TYPE FP_TO_INT_D(__FP_D_TYPE X, _FP_I_TYPE rsz, _FP_I_TYPE rsigned)
{
   return _FP_TO_INT('d',__FP_D_SIZE,(_FP_VAL)X,rsz,rsigned);
}

INLINE void FP_FROM_INT_D(__FP_D_TYPE X, _FP_CONV_I_TYPE r, _FP_I_TYPE rs)
{
   _FP_FROM_INT('d',__FP_D_SIZE,(_FP_VAL)X,r,rs);
}

INLINE void FP_CONV_S_D(__FP_S_TYPE X, __FP_D_TYPE Y)
{
   FP_CONV('s', 'd', __FP_S_SIZE, __FP_D_SIZE, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_CONV_Q_D(__FP_Q_TYPE X, __FP_D_TYPE Y)
{
   FP_CONV('q', 'd', __FP_Q_SIZE, __FP_D_SIZE, (_FP_VAL)X, (_FP_VAL)Y);
}

#endif /* not DOUBLE_H */
