/*
 * Basic one-word fraction declaration and manipulation.
 */

/* FP variables are really pointers to structs: */
typedef struct 
{
   _FP_COMMON_DECL;                               /* from op-common-pre.h */
   _FP_W_TYPE f;
} _FP_1_struct;

typedef _FP_1_struct* _FP_1;

INLINE void _FP_FRAC_COPY_1(_FP_1 D, _FP_1 S)
{
   D->f = S->f;
}

INLINE _FP_W_TYPE* _FP_FRAC_HIGH_1(_FP_1 X)
{
   return &(X->f);
}

INLINE _FP_W_TYPE* _FP_FRAC_LOW_1(_FP_1 X)
{
   return &(X->f);
}

INLINE _FP_W_TYPE* _FP_FRAC_WORD_1(_FP_1 X, _FP_I_TYPE w)
{
   return &(X->f);
}

INLINE void _FP_FRAC_SET_ZERO_1(_FP_1 X)
{
   X->f = 0;
}

INLINE void _FP_FRAC_SET_MIN_1(_FP_1 X)
{
   X->f = 1;
}

INLINE void _FP_FRAC_SET_NAN_1(char fs, _FP_1 X)
{
   X->f = _FP_NANFRAC(fs,0);
}

INLINE void _FP_FRAC_ADDI_1(_FP_1 X, _FP_W_TYPE I)
{
   X->f += I;
}

INLINE void _FP_FRAC_SLL_1(_FP_1 X, _FP_I_TYPE N)
{
   /* This code appears to make interesting assumptions 
    * about whether X += X is more efficient than X<<=N
    * and whether the compiler is intelligent enough
    * to optimise things. Let it stand for the moment -- PMM 
    */
   if (__builtin_constant_p(N) && N == 1)
      X->f += X->f;
   else
      X->f <<= N;
}

INLINE void _FP_FRAC_SRL_1(_FP_1 X, _FP_I_TYPE N)
{
   X->f >>= N;
}

/* op-common-post.h also uses this, sigh */
INLINE _FP_W_TYPE __FP_FRAC_SRS_1(_FP_W_TYPE X, _FP_I_TYPE N, _FP_I_TYPE sz)
{
   return (X >> N | (__builtin_constant_p(N) && N == 1
                     ? X & 1 : (X << (_FP_W_TYPE_SIZE - N)) != 0));
}

/* Right shift with sticky-lsb.  */
INLINE void _FP_FRAC_SRS_1(_FP_1 X, _FP_I_TYPE N, _FP_I_TYPE sz)
{
   X->f = __FP_FRAC_SRS_1(X->f,N,sz);
}

INLINE void _FP_FRAC_ADD_1(_FP_1 R, _FP_1 X, _FP_1 Y)
{
   R->f = X->f + Y->f;
}

INLINE void _FP_FRAC_SUB_1(_FP_1 R, _FP_1 X, _FP_1 Y)
{
   R->f = X->f - Y->f;
}

INLINE _FP_I_TYPE _FP_FRAC_CLZ_1(_FP_1 X)
{
   return __FP_CLZ(X->f);
}

/* Predicates */
INLINE _FP_I_TYPE _FP_FRAC_NEGP_1(_FP_1 X)
{
   return ((_FP_WS_TYPE)X->f < 0);
}

INLINE _FP_I_TYPE _FP_FRAC_ZEROP_1(_FP_1 X)
{
   return (X->f == 0);
}

INLINE _FP_I_TYPE _FP_FRAC_OVERP_1(char fs, _FP_1 X)
{
   return (X->f & _FP_OVERFLOW(fs));
}

INLINE _FP_I_TYPE _FP_FRAC_EQ_1(_FP_1 X, _FP_1 Y)
{
   return (X->f == Y->f);
}

INLINE _FP_I_TYPE _FP_FRAC_GE_1(_FP_1 X, _FP_1 Y)
{
   return (X->f >= Y->f);
}

INLINE _FP_I_TYPE _FP_FRAC_GT_1(_FP_1 X, _FP_1 Y)
{
   return (X->f > Y->f);
}

/*
 * Unpack the raw bits of a native fp value.  Do not classify or
 * normalize the data. 
 * Added bugfix cast from sparc64's sfp-machine.h -- PMM 02/1998
 */
INLINE void _FP_UNPACK_RAW_1(char fs, _FP_1 X, void *val)
{
   if (fs == 's')
   {
#ifdef __FP_S_IS_1_WORD
      union _FP_UNION_S *flo = (union _FP_UNION_S *) val;
      X->f = flo->bits.frac;
      X->e = flo->bits.exp;
      X->s = flo->bits.sign;
#endif
   }
   else if (fs == 'd')
   {
#ifdef __FP_D_IS_1_WORD
      union _FP_UNION_D *flo = (union _FP_UNION_D *) val;
      X->f = flo->bits.frac;
      X->e = flo->bits.exp;
      X->s = flo->bits.sign;
#endif
   }
   else if (fs == 'q')
   {
#ifdef __FP_Q_IS_1_WORD
      union _FP_UNION_Q *flo = (union _FP_UNION_Q *) val;
      X->f = flo->bits.frac;
      X->e = flo->bits.exp;
      X->s = flo->bits.sign;
#endif
   }
}


/*
 * Repack the raw bits of a native fp value.
 */
INLINE void _FP_PACK_RAW_1(char fs, void *val, _FP_1 X)
{
   if (fs == 's')
   {
#ifdef __FP_S_IS_1_WORD
      union _FP_UNION_S *flo = (union _FP_UNION_S *) val;
      flo->bits.frac = X->f;
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
   else if (fs == 'd')
   {
#ifdef __FP_D_IS_1_WORD
      union _FP_UNION_D *flo = (union _FP_UNION_D *) val;
      flo->bits.frac = X->f;
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
   else if (fs == 'q')
   {
#ifdef __FP_Q_IS_1_WORD
      union _FP_UNION_Q *flo = (union _FP_UNION_Q *) val;
      flo->bits.frac = X->f;
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
}

/*
 * Multiplication algorithms:
 */

/* Basic.  Assuming the host word size is >= 2*FRACBITS, we can do the
   multiplication immediately.  */
INLINE void _FP_MUL_MEAT_1_imm(char fs, _FP_1 R, _FP_1 X, _FP_1 Y)
{
   R->f = X->f * Y->f;
   /* Normalize; since we know where the msb of the multiplicands
    * were (bit B), we know that the msb of the of the product is
    * at either 2B or 2B-1.
    */
   _FP_FRAC_SRS_1(R, _FP_WFRACBITS(fs) - 1, 2 * _FP_WFRACBITS(fs));
}

/* Given a 1W * 1W => 2W primitive, do the extended multiplication.  */
#ifndef _FP_MUL_MEAT_1_wide_doit
#define _FP_MUL_MEAT_1_wide_doit(a,b,c,d) /* this routine not used */
#endif
INLINE void _FP_MUL_MEAT_1_wide(char fs, _FP_1 R, _FP_1 X, _FP_1 Y)
{
   _FP_DECL(_FP_2, Z);
   _FP_MUL_MEAT_1_wide_doit(Z->f1, Z->f0, X->f, Y->f);
   /* Normalize; since we know where the msb of the multiplicands
    * were (bit B), we know that the msb of the of the product is
    * at either 2B or 2B-1. 
    */
   _FP_FRAC_SRS_2(Z, _FP_WFRACBITS(fs)-1, 2*_FP_WFRACBITS(fs));
   R->f = Z->f0;
}

/* Finally, a simple widening multiply algorithm.  What fun!  */
INLINE void _FP_MUL_MEAT_1_hard(char fs, _FP_1 R, _FP_1 X, _FP_1 Y)
{
   _FP_W_TYPE xh, xl, yh, yl;
   _FP_DECL(_FP_2, z);
   _FP_DECL(_FP_2, a);
   
   /* split the words in half */
   xh = X->f >> (_FP_W_TYPE_SIZE / 2);
   xl = X->f & (((_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE/2)) - 1);
   yh = Y->f >> (_FP_W_TYPE_SIZE / 2);
   yl = Y->f & (((_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE/2)) - 1);
 
   /* multiply the pieces */
   z->f0 = xl * yl;
   a->f0 = xh * yl;
   a->f1 = xl * yh;
   z->f1 = xh * yh;
   
   /* reassemble into two full words */
   if ((a->f0 += a->f1) < a->f1)
       z->f1 += (_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE/2);
   a->f1 = a->f0 >> (_FP_W_TYPE_SIZE / 2);
   a->f0 = a->f0 << (_FP_W_TYPE_SIZE / 2);
   _FP_FRAC_ADD_2(z,z,a);
   
   /* normalize */
   _FP_FRAC_SRS_2(z, _FP_WFRACBITS(fs)-1, 2*_FP_WFRACBITS(fs));
   R->f = z->f0;
}

/*
 * Division algorithms:
 */

/* Basic.  Assuming the host word size is >= 2*FRACBITS, we can do the
 * division immediately.  Give this macro either _FP_DIV_HELP_imm for
 * C primitives or _FP_DIV_HELP_ldiv for the ISO function.  Which you
 * choose will depend on what the compiler does with divrem4.  
 * NB: 'give' here means defining _FP_DIV_MEAT_1_imm_doit -- PMM
 */
#ifndef _FP_DIV_MEAT_1_imm_doit
#define _FP_DIV_MEAT_1_imm_doit(a,b,c,d)
#endif
INLINE void _FP_DIV_MEAT_1_imm(char fs, _FP_1 R, _FP_1 X, _FP_1 Y)
{
   _FP_W_TYPE q, r;
   /* this next line is dire but it's how the macro did it :->   -- PMM */
   X->f <<= (X->f < Y->f ? R->e--, _FP_WFRACBITS(fs) : _FP_WFRACBITS(fs)-1);
   _FP_DIV_MEAT_1_imm_doit(q,r, X->f, Y->f);
   R->f = q | (r != 0);
}

/* GCC's longlong.h defines a 2W / 1W => (1W,1W) primitive udiv_qrnnd
   that may be useful in this situation.  This first is for a primitive
   that requires normalization, the second for one that does not.  Look
   for UDIV_NEEDS_NORMALIZATION to tell which your machine needs.  */

INLINE void _FP_DIV_MEAT_1_udiv_norm(char fs, _FP_1 R, _FP_1 X, _FP_1 Y)
{
   _FP_W_TYPE nh, nl, q, r;
   
   /* normalize Y -- i.e. make the most significant bit set */
   Y->f <<= _FP_WFRACXBITS(fs)-1;
   
   /* shift X op correspondingly high, that is, up one full word. */
   if (X->f <= Y->f)
   {
      nl = 0;
      nh = X->f;
   }
   else
   {
      R->e++;
      nl = X->f << (_FP_W_TYPE_SIZE - 1);
      nh = X->f >> 1;
   }
   
   udiv_qrnnd(q,r,nh,nl,Y->f);
   R->f = q | (r != 0);
}

INLINE void _FP_DIV_MEAT_1_udiv(char fs, _FP_1 R, _FP_1 X, _FP_1 Y)
{
   _FP_W_TYPE nh, nl, q, r;
   if (X->f < Y->f)
   {
      R->e--;
      nl = X->f << _FP_WFRACBITS(fs);
      nh = X->f >> _FP_WFRACXBITS(fs);
   }
   else
   {
      nl = X->f << (_FP_WFRACBITS(fs) - 1);
      nh = X->f >> (_FP_WFRACXBITS(fs) + 1);
   }
   udiv_qrnnd(q, r, nh, nl, Y->f);
   R->f = q | (r != 0);
}

/*
 * Square root algorithms:
 * We have just one right now, maybe Newton approximation
 * should be added for those machines where division is fast.
 */
INLINE void _FP_SQRT_MEAT_1(_FP_1 R, _FP_1 X, _FP_W_TYPE q)
{
   _FP_W_TYPE s = 0, t;
   
   while (q)
   {
      t = s + q;
      if (t <= X->f)
      {
         s = t + q;
         X->f -= t;
         R->f += q;
      }
      _FP_FRAC_SLL_1(X,1);
      q >>= 1;
   }
}


/*
 * Assembly/disassembly for converting to/from integral types.  
 * No shifting or overflow handled here.
 */

INLINE unsigned _FP_CONV_I_TYPE _FP_FRAC_ASSEMBLE_1(_FP_1 X, _FP_I_TYPE rsize)
{
   return X->f;
}

INLINE void _FP_FRAC_DISASSEMBLE_1(_FP_1 X, unsigned _FP_CONV_I_TYPE r, _FP_I_TYPE rsize)
{
   X->f = r;
}

/*
 * Convert FP values between word sizes. These macros go in op-n.h
 * where n is the smaller of the two word sizes involved.
 */
INLINE void _FP_FRAC_CONV_1_1(char dfs, char sfs, _FP_1 D, _FP_1 S)
{
   D->f = S->f;
   if (_FP_WFRACBITS(sfs) > _FP_WFRACBITS(dfs))
      _FP_FRAC_SRS_1(D, _FP_WFRACBITS(sfs) - _FP_WFRACBITS(dfs), _FP_WFRACBITS(sfs));
   else
      D->f <<= _FP_WFRACBITS(dfs) - _FP_WFRACBITS(sfs);
}

INLINE void _FP_FRAC_CONV_4_1(char dfs, char sfs, _FP_4 D, _FP_1 S)
{
   D->f[0] = S->f;
   D->f[1] = D->f[2] = D->f[3] = 0;
   _FP_FRAC_SLL_4(D, _FP_WFRACBITS(dfs) - _FP_WFRACBITS(sfs));
}

INLINE void _FP_FRAC_CONV_1_4(char dfs, char sfs, _FP_1 D, _FP_4 S)
{
   _FP_FRAC_SRS_4(S, _FP_WFRACBITS(sfs) - _FP_WFRACBITS(dfs), _FP_WFRACBITS(sfs));
   D->f = S->f[0];
}

INLINE void _FP_FRAC_CONV_1_2(char dfs, char sfs, _FP_1 D, _FP_2 S)
{
   _FP_FRAC_SRS_2(S, _FP_WFRACBITS(sfs) - _FP_WFRACBITS(dfs), _FP_WFRACBITS(sfs));
   D->f = S->f0;
}

INLINE void _FP_FRAC_CONV_2_1(char dfs, char sfs, _FP_2 D, _FP_1 S)
{
   D->f0 = S->f;
   D->f1 = 0;
   _FP_FRAC_SLL_2(D, _FP_WFRACBITS(dfs) - _FP_WFRACBITS(sfs));
}
