/*
 * Basic two-word fraction declaration and manipulation.
 */

/* FP variables are really pointers to structs: */
typedef struct 
{
   _FP_COMMON_DECL;                               /* from op-common-pre.h */
   _FP_W_TYPE f0;
   _FP_W_TYPE f1;
} _FP_2_struct;

typedef _FP_2_struct* _FP_2;

INLINE void _FP_FRAC_COPY_2(_FP_2 D, _FP_2 S)
{
   D->f0 = S->f0;
   D->f1 = S->f1;
}

INLINE _FP_W_TYPE* _FP_FRAC_HIGH_2(_FP_2 X)
{
   return &(X->f1);
}

INLINE _FP_W_TYPE* _FP_FRAC_LOW_2(_FP_2 X)
{
   return &(X->f0);
}

INLINE _FP_W_TYPE* _FP_FRAC_WORD_2(_FP_2 X, _FP_I_TYPE w)
{
   return (w == 0) ? &(X->f0) : &(X->f1);
}

/* we shouldn't use this; use _FP_FRAC_SET_NAN() and co instead...
INLINE void _FP_FRAC_SET_2(_FP_2 X, _FP_W_TYPE I0, _FP_W_TYPE I1)
{
   X->f0 = I0;
   X->f1 = I1;
}
*/

INLINE void _FP_FRAC_SET_ZERO_2(_FP_2 X)
{
   X->f0 = X->f1 = 0;
}

INLINE void _FP_FRAC_SET_MIN_2(_FP_2 X)
{
   X->f0 = 0;
   X->f1 = 1;
}

INLINE void _FP_FRAC_SET_NAN_2(char fs, _FP_2 X)
{
   /* _FP_NANFRAC is machine dependent, defined in sfp-machine.h */
   X->f0 = _FP_NANFRAC(fs,0);
   X->f1 = _FP_NANFRAC(fs,1);
}

INLINE void _FP_FRAC_SLL_2(_FP_2 X, _FP_I_TYPE N)
{
   if (N < _FP_W_TYPE_SIZE)
   {
      if (__builtin_constant_p(N) && N == 1)      /* This is an optimisation: << 1 is easier. */
      {                                           /* __builtin_constant_p() means that only one arm of the  */
                                                  /* if() is ever compiled in for a given use of the function. */
         X->f1 = X->f1 + X->f1 + (((_FP_WS_TYPE) X->f0) < 0);
         X->f0 += X->f0;
      }
      else
      {
         X->f1 = X->f1 << N | X->f0 >> (_FP_W_TYPE_SIZE - N);
         X->f0 <<= N;
      }
   }
   else
   {
      X->f1 = X->f0 << (N - _FP_W_TYPE_SIZE);
      X->f0 = 0;
   }
}

INLINE void _FP_FRAC_SRL_2(_FP_2 X, _FP_I_TYPE N)
{
   if (N < _FP_W_TYPE_SIZE)
   {
      X->f0 = X->f0 >> N | X->f1 << (_FP_W_TYPE_SIZE - N);
      X->f1 >>= N;
   }
   else
   {
      X->f0 = X->f1 >> (N - _FP_W_TYPE_SIZE);
      X->f1 = 0;
   }
}

/* Right shift with sticky-lsb.  */
INLINE void _FP_FRAC_SRS_2(_FP_2 X, _FP_I_TYPE N, _FP_I_TYPE sz)
{
   if (N < _FP_W_TYPE_SIZE)
   {
      X->f0 = (X->f1 << (_FP_W_TYPE_SIZE - N) 
               | X->f0 >> N
               | ((__builtin_constant_p(N) && N == 1) 
                 ? (X->f0 & 1) 
                 : ((X->f0 << (_FP_W_TYPE_SIZE - N)) != 0)));
      X->f1 >>= N;
   }
   else
   {
      X->f0 = (X->f1 >> (N - _FP_W_TYPE_SIZE) |
              (((X->f1 << (2*_FP_W_TYPE_SIZE - N)) | X->f0) != 0));
      /* '2*_FP_W_TYPE_SIZE' used to be 'sz' -- PMM 04/1998 */
      X->f1 = 0;
   }
}

INLINE void _FP_FRAC_ADDI_2(_FP_2 X, _FP_W_TYPE I)
{
   /* equivalent to:  X->f1 += ((X->f0 += i) < i); */
   add_ssaaaa(X->f1,X->f0,X->f1,X->f0,0,I);
}

INLINE void _FP_FRAC_ADD_2(_FP_2 R, _FP_2 X, _FP_2 Y)
{
   /* equivalent to:  R->f1 = X->f1 + Y->f1 + ((R->f0 = X->f0 + Y->f0) < X->f0) */
   add_ssaaaa(R->f1,R->f0,X->f1,X->f0,Y->f1,Y->f0);
}

INLINE void _FP_FRAC_SUB_2(_FP_2 R, _FP_2 X, _FP_2 Y)
{
   /* equivalent to:  (rh = xh - yh - ((rl = xl - yl) > xl)) */
   sub_ddmmss(R->f1,R->f0,X->f1,X->f0,Y->f1,Y->f0);
}

INLINE void __FP_FRAC_SUB_2(_FP_W_TYPE* r1, _FP_W_TYPE* r0, 
                            _FP_W_TYPE x1, _FP_W_TYPE x0, _FP_W_TYPE y1, _FP_W_TYPE y0)
{
   /* routine used only by functions in this file. As above, but works on
    * split fractional parts rather than complete _FP_2 objects.
    */
   sub_ddmmss(*r1,*r0,x1,x0,y1,y0);
}

/* this ought to be internal but _FP_FROM_INT in op-common-post.h uses it :-< */
INLINE _FP_I_TYPE __FP_CLZ_2(_FP_W_TYPE h, _FP_W_TYPE l)
{
   if (h)
      return __FP_CLZ(h);
   else
      return _FP_W_TYPE_SIZE + __FP_CLZ(l);
}

INLINE _FP_I_TYPE _FP_FRAC_CLZ_2(_FP_2 X)
{
   return __FP_CLZ_2(X->f1, X->f0);
}

/* Predicates */
INLINE _FP_I_TYPE _FP_FRAC_NEGP_2(_FP_2 X)
{
   return ((_FP_WS_TYPE)X->f1 < 0);
}

INLINE _FP_I_TYPE _FP_FRAC_ZEROP_2(_FP_2 X)
{
   return ((X->f1 | X->f0) == 0);
}

INLINE _FP_I_TYPE _FP_FRAC_OVERP_2(char fs, _FP_2 X)
{
   return (X->f1 & _FP_OVERFLOW(fs));
}

INLINE _FP_I_TYPE _FP_FRAC_EQ_2(_FP_2 X, _FP_2 Y)
{
   return (X->f1 == Y->f1 && X->f0 == Y->f0);
}

INLINE _FP_I_TYPE _FP_FRAC_GT_2(_FP_2 X, _FP_2 Y)
{
   return (X->f1 > Y->f1 || (X->f1 == Y->f1 && X->f0 > Y->f0));
}

INLINE _FP_I_TYPE _FP_FRAC_GE_2(_FP_2 X, _FP_2 Y)
{
   return (X->f1 > Y->f1 || (X->f1 == Y->f1 && X->f0 >= Y->f0));
}

/*
 * Internals 
 */

/*
 * Unpack the raw bits of a native fp value.  Do not classify or
 * normalize the data.
 * Bugfix casts from sparc64 sfp-machine.h added -- PMM 02/1998
 * The if()..else if()..else structure is ugly but unavoidable, I think.
 * The #ifdefs are because otherwise we're trying to access elements
 * of unions like _FP_UNION_S that don't exist. The compiler can't
 * tell that we never call this with fs=='s' unless _FP_UNION_S
 * actually has the right elements, so we need to get cpp to
 * edit the references out. Ugly, isn't it?
 */

INLINE void _FP_UNPACK_RAW_2(char fs, _FP_2 X, void *val)
{
   if (fs == 's')
   {
#ifdef __FP_S_IS_2_WORDS
      union _FP_UNION_S *flo = (union _FP_UNION_S *) val;
      X->f0 = flo->bits.frac0;
      X->f1 = flo->bits.frac1;
      X->e  = flo->bits.exp;
      X->s  = flo->bits.sign;
#endif
   }
   else if (fs == 'd')
   {
#ifdef __FP_D_IS_2_WORDS
      union _FP_UNION_D *flo = (union _FP_UNION_D *) val;
      X->f0 = flo->bits.frac0;
      X->f1 = flo->bits.frac1;
      X->e  = flo->bits.exp;
      X->s  = flo->bits.sign;
#endif
   }
   else if (fs == 'q')
   {
#ifdef __FP_Q_IS_2_WORDS
      union _FP_UNION_Q *flo = (union _FP_UNION_Q *) val;
      X->f0 = flo->bits.frac0;
      X->f1 = flo->bits.frac1;
      X->e  = flo->bits.exp;
      X->s  = flo->bits.sign;
#endif
   }
}

/*
 * Repack the raw bits of a native fp value.
 */
INLINE void _FP_PACK_RAW_2(char fs, void *val, _FP_2 X)
{
   if (fs == 's')
   {
#ifdef __FP_S_IS_2_WORDS
      union _FP_UNION_S *flo = (union _FP_UNION_S *) val;
      flo->bits.frac0 = X->f0;
      flo->bits.frac1 = X->f1;
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
   else if (fs == 'd')
   {
#ifdef __FP_D_IS_2_WORDS
      union _FP_UNION_D *flo = (union _FP_UNION_D *) val;
      flo->bits.frac0 = X->f0;
      flo->bits.frac1 = X->f1;
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
   else
   {
#ifdef __FP_Q_IS_2_WORDS
      union _FP_UNION_Q *flo = (union _FP_UNION_Q *) val;
      flo->bits.frac0 = X->f0;
      flo->bits.frac1 = X->f1;
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
}

/*
 * Multiplication algorithms:
 */

/* Given a 1W * 1W => 2W primitive, do the extended multiplication.  */
#ifndef _FP_MUL_MEAT_2_wide_doit
#define _FP_MUL_MEAT_2_wide_doit(a,b,c,d) /* do nothing */
#endif
INLINE void _FP_MUL_MEAT_2_wide(char fs, _FP_2 R, _FP_2 X, _FP_2 Y)
{
   _FP_DECL(_FP_4,z);
   _FP_DECL(_FP_4,b);
   _FP_DECL(_FP_4,c);
   
   _FP_FRAC_SET_ZERO_4(b);
   _FP_FRAC_SET_ZERO_4(c);
   
   _FP_MUL_MEAT_2_wide_doit(*_FP_FRAC_WORD_4(z,1), *_FP_FRAC_WORD_4(z,0), X->f0 ,Y->f0);
   _FP_MUL_MEAT_2_wide_doit(*_FP_FRAC_WORD_4(b,2), *_FP_FRAC_WORD_4(b,1), X->f0, Y->f1);
   _FP_MUL_MEAT_2_wide_doit(*_FP_FRAC_WORD_4(c,2), *_FP_FRAC_WORD_4(c,1), X->f1, Y->f0);
   _FP_MUL_MEAT_2_wide_doit(*_FP_FRAC_WORD_4(z,3), *_FP_FRAC_WORD_4(z,2), X->f1, Y->f1);
   
   _FP_FRAC_ADD_4(z,z,b);
   _FP_FRAC_ADD_4(z,z,c);
   
   /* Normalize; since we know where the msb of the multiplicands were (bit B), 
    * we know that the msb of the of the product is at either 2B or 2B-1.  
    */
   _FP_FRAC_SRS_4(z, _FP_WFRACBITS(fs)-1, 2*_FP_WFRACBITS(fs));
   R->f0 = *_FP_FRAC_WORD_4(z,0);
   R->f1 = *_FP_FRAC_WORD_4(z,1);
}

#ifdef UNDEF
/* This next macro appears to be totally broken. Fortunately nowhere
 * seems to use it :-> The problem is that we define _z[4] but
 * then use it in _FP_FRAC_SRS_4, which will attempt to access
 * _z_f[n] which will cause an error. The fix probably involves 
 * declaring it with _FP_FRAC_DECL_4, see previous macro. -- PMM 02/1998 
 *
 * I've done a partial translation to inline functions; fix it if you
 * want to use it! The original macro is after the partial translation. -- PMM 03/1998
 */
INLINE void _FP_MUL_MEAT_2_gmp(char fs, _FP_2 R, _FP_2 X, _FP_2 Y)
{
   _FP_W_TYPE x[2], y[2];
   _FP_DECL(_FP_4,z);
   x[0] = X->f0;
   x[1] = X->f1;
   y[0] = Y->f0;
   y[1] = Y->f1;
  
   /* This should probably not pass z directly, but expand it 
    * into individual words [using z->f is cheating :-> ]
    */
   mpn_mul_n(z,x,y,2);
   
   /* Normalize; since we know where the msb of the multiplicands were (bit B), 
    * we know that the msb of the of the product is at either 2B or 2B-1.  
    */
   _FP_FRAC_SRS_4(z, _FP_WFRACBITS(fs)-1, 2*_FP_WFRACBITS(fs));
   R->f0 = _FP_FRAC_WORD_4(z,0);
   R->f1 = _FP_FRAC_WORD_4(z,1);
}

#define _FP_MUL_MEAT_2_gmp(fs, R, X, Y)					\
  do {									\
    _FP_W_TYPE _x[2], _y[2], _z[4];					\
    _x[0] = X##_f0; _x[1] = X##_f1;					\
    _y[0] = Y##_f0; _y[1] = Y##_f1;					\
									\
    mpn_mul_n(_z, _x, _y, 2);						\
									\
    /* Normalize since we know where the msb of the multiplicands	\
       were (bit B), we know that the msb of the of the product is	\
       at either 2B or 2B-1.  */					\
    _FP_FRAC_SRS_4(_z, _FP_WFRACBITS##_fs-1, 2*_FP_WFRACBITS_##fs);	\
    R##_f0 = _z[0];							\
    R##_f1 = _z[1];							\
  } while (0)
#endif /* UNDEF */

/*
 * Division algorithms:
 * Look, NetBSD seems to be able to comment algorithms. Can't you?
 * This now appears to work, but I still don't really know why.
 */
INLINE void _FP_DIV_MEAT_2_udiv_64(char fs, _FP_2 R, _FP_2 X, _FP_2 Y)
{
   /* putting this declaration here is a bit vile... */
   extern void _fp_udivmodti4(_FP_W_TYPE q[2], _FP_W_TYPE r[2],
                              _FP_W_TYPE n1, _FP_W_TYPE n0,
                              _FP_W_TYPE d1, _FP_W_TYPE d0);
   _FP_W_TYPE n_f3, n_f2, n_f1, n_f0, q_f1, q_f0, rmem[2], qmem[2];
   _FP_DECL(_FP_2,m); _FP_DECL(_FP_2,r);
   
   /* I think this check is to ensure that the result is normalised.   
    * Assuming X,Y normalised (ie in [1.0,2.0)) X/Y will be in         
    * [0.5,2.0). Furthermore, it will be less than 1.0 iff X < Y.      
    * In this case we tweak things. (this is based on comments in      
    * the NetBSD FPU emulation code. )                                 
    * We know X,Y are normalised because we ensure this as part of     
    * the unpacking process. -- PMM                                  
    */
    if (_FP_FRAC_GE_2(X,Y))                       /* was 'GT' -- PMM 04/1998 */
    {
       /* R->e++; */
       n_f3 = X->f1 >> 1;
       n_f2 = X->f1 << (_FP_W_TYPE_SIZE - 1) | X->f0 >> 1;
       n_f1 = X->f0 << (_FP_W_TYPE_SIZE - 1);
       n_f0 = 0;
    }
    else
    {
       R->e--;
       n_f3 = X->f1;
       n_f2 = X->f0;
       n_f1 = n_f0 = 0;
    }
    /* Normalize, i.e. make the most significant bit of the
     * denominator set. 
     */
    _FP_FRAC_SLL_2(Y, _FP_WFRACXBITS(fs));
    
    /* Do the 256/128 bit division given the 128-bit _fp_udivmodtf4
     * primitive snagged from libgcc2.c.  
     */
    _fp_udivmodti4(qmem, rmem, n_f3, n_f2, 0, Y->f1);
    q_f1 = qmem[0];
    umul_ppmm(m->f1, m->f0, q_f1, Y->f0);
    r->f1 = rmem[0];
    r->f0 = n_f1;
    
    if (_FP_FRAC_GT_2(m,r))
    {
       q_f1--;
       _FP_FRAC_ADD_2(r,r,Y);
       if (_FP_FRAC_GE_2(r, Y) && _FP_FRAC_GT_2(m, r))
       {
          q_f1--;
          _FP_FRAC_ADD_2(r,r,Y);
       }
    }
    _FP_FRAC_SUB_2(r,r,m);
    
    _fp_udivmodti4(qmem, rmem, r->f1, r->f0, 0, Y->f1);
    q_f0 = qmem[0];
    umul_ppmm(m->f1, m->f0, q_f0, Y->f0);
    r->f1 = rmem[0];
    r->f0 = n_f0;
    if (_FP_FRAC_GT_2(m,r))
    {
       q_f0--;
       _FP_FRAC_ADD_2(r,r,Y);
       if (_FP_FRAC_GE_2(r,Y) && _FP_FRAC_GT_2(m,r))
       {
          q_f0--;
          _FP_FRAC_ADD_2(r,r,Y);
       }
    }
    _FP_FRAC_SUB_2(r,r,m);
   
    R->f1 = q_f1;
    R->f0 = q_f0 | ((r->f1 | r->f0) != 0);
    /* adjust so answer is normalized again. I'm not sure what the 	
     * final sz param should be. In practice it's never used by any SRS
     * (I think it's an artifact and that parameter should probably be 
     * removed from the SRS functions...)
     */	
/*     _FP_FRAC_SRS_2(R,1,_FP_WFRACBITS(fs)); */
}

#ifndef mpn_divrem
#define mpn_divrem(A,B,C,D,E,F) /* not used */
#endif
/* this macro can't have ever been tested because it used 
 * _FP_WFRACBITS rather than _FP_WFRACBITS_##fs. I've corrected
 * this obvious error whilst translating to inline functions;
 * more subtle bugs may lurk undiscovered -- PMM 03/1998.
 */
INLINE void _FP_DIV_MEAT_2_gmp(char fs, _FP_2 R, _FP_2 X, _FP_2 Y)
{
   _FP_W_TYPE x[4], y[2], z[4];
   y[0] = Y->f0;
   y[1] = Y->f1;
   x[0] = x[3] = 0;
   if (_FP_FRAC_GT_2(X,Y))
   {
      R->e++;
      x[1] = (X->f0 << (_FP_WFRACBITS(fs)-1 - _FP_W_TYPE_SIZE) 
              | X->f1 >> (_FP_W_TYPE_SIZE - (_FP_WFRACBITS(fs)-1 - _FP_W_TYPE_SIZE)));
      x[2] = X->f1 << (_FP_WFRACBITS(fs)-1 - _FP_W_TYPE_SIZE);
   }
   else
   {
      x[1] = (X->f0 << (_FP_WFRACBITS(fs) - _FP_W_TYPE_SIZE) 
              | X->f1 >> (_FP_W_TYPE_SIZE - (_FP_WFRACBITS(fs) - _FP_W_TYPE_SIZE)));
      x[2] = X->f1 << (_FP_WFRACBITS(fs) - _FP_W_TYPE_SIZE);
   }

   /* (void) deleted -- PMM */
   mpn_divrem(z,0,x,4,y,2);
   R->f1 = z[1];
   R->f0 = z[0] | ((x[0] | x[1]) != 0);
}


/*
 * Square root algorithms:
 * We have just one right now, maybe Newton approximation
 * should be added for those machines where division is fast.
 * 
 * This has been modified from the macro version. We now define and initialise
 * S and T inside this function rather than in _FP_SQRT_MEAT in op-common-post.h
 * [both should effectively be _FP_n variables. S must be initialised to zero.]
 */
INLINE void _FP_SQRT_MEAT_2(_FP_2 R, _FP_2 X, _FP_W_TYPE q)
{
   _FP_W_TYPE tf1, tf0, sf1 = 0, sf0 = 0;
   
   while (q)
   {
      tf1 = sf1 + q;
      if (tf1 <= X->f1)
      {
         sf1 = tf1 + q;
         X->f1 -= tf1;
         R->f1 += q;
      }
      _FP_FRAC_SLL_2(X,1);
      q >>= 1;
   }
   q = (_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE - 1);
   while (q)
   {
      tf0 = sf0 + q;
      tf1 = sf1;
      /* Is this right? Shouldn't it be 'if(T <= X)'? (ie last '<' should be '<=')? -- PMM */
      if (tf1 < X->f1 || (tf1 == X->f1 && tf0 < X->f0))
      {
         sf0 = tf0 + q;
         if (((_FP_WS_TYPE)tf0) < 0 && ((_FP_WS_TYPE)sf0) >= 0)
            sf1++;
         __FP_FRAC_SUB_2(&(X->f1),&(X->f0),X->f1,X->f0,tf1,tf0);
         R->f0 += q;
      }
      _FP_FRAC_SLL_2(X,1);
      q >>= 1;
   }
}


/*
 * Assembly/disassembly for converting to/from integral types.  
 * No shifting or overflow handled here.
 */
INLINE unsigned _FP_CONV_I_TYPE _FP_FRAC_ASSEMBLE_2(_FP_2 X, _FP_I_TYPE rsize)
{
   if (rsize <= _FP_W_TYPE_SIZE)
      return X->f0;
   else
   {
#if _FP_CONV_I_TYPE_SIZE > _FP_W_TYPE_SIZE
      _FP_W_TYPE r = X->f1;
      r <<= _FP_W_TYPE_SIZE;
      r += X->f0;
      return r;
#else
      /* can't happen unless caller is buggy */
      printk("_FP_FRAC_ASSEMBLE_2: bogus rsize (bug!)\n");
      return 0;
#endif
   }
}


INLINE void _FP_FRAC_DISASSEMBLE_2(_FP_2 X, unsigned _FP_CONV_I_TYPE r, _FP_I_TYPE rsize)
{
   X->f0 = r;
   if (rsize <= _FP_W_TYPE_SIZE)
      X->f1 = 0;
   else
   {
#if _FP_CONV_I_TYPE_SIZE > _FP_W_TYPE_SIZE
      X->f1 = r >> _FP_W_TYPE_SIZE;
#else
      /* can't happen unless caller is buggy */
      printk("_FP_FRAC_ASSEMBLE_2: bogus rsize (bug!)\n");
#endif
   }
}


/* Convert FP values between word sizes. This appears to be more
 * complicated than I'd have expected it to be, so these might be
 * wrong... These macros are in any case somewhat bogus because they
 * use information about what various FRAC_n variables look like 
 * internally [eg, that 2 word vars are X_f0 and x_f1]. But so do
 * the ones I didn't write  -- PMM
 */
INLINE void _FP_FRAC_CONV_2_4(char dfs, char sfs, _FP_2 D, _FP_4 S)
{
   _FP_FRAC_SRS_4(S, _FP_WFRACBITS(sfs) - _FP_WFRACBITS(dfs), _FP_WFRACBITS(sfs));
   D->f0 = S->f[0];
   D->f1 = S->f[1];
}

INLINE void _FP_FRAC_CONV_4_2(char dfs, char sfs, _FP_4 D, _FP_2 S)
{
   D->f[0] = S->f0;
   D->f[1] = S->f1;
   D->f[2] = D->f[3] = 0;
   _FP_FRAC_SLL_4(D, _FP_WFRACBITS(dfs) - _FP_WFRACBITS(sfs));
}

/* there wasn't a 2->2 conversion in the original macro version;
 * I've added it for completeness  -- PMM 04/1998
 */
INLINE void _FP_FRAC_CONV_2_2(char dfs, char sfs, _FP_2 D, _FP_2 S)
{
   D->f0 = S->f0; D->f1 = S->f1;
   if (_FP_WFRACBITS(sfs) > _FP_WFRACBITS(dfs))
      _FP_FRAC_SRS_2(D, _FP_WFRACBITS(sfs) - _FP_WFRACBITS(dfs), _FP_WFRACBITS(sfs)); 
   else if (_FP_WFRACBITS(sfs) < _FP_WFRACBITS(dfs))
      _FP_FRAC_SLL_2(D, _FP_WFRACBITS(dfs) - _FP_WFRACBITS(sfs));
}
