/* these were in soft-fp.h, but can be put here now: */

INLINE void _FP_ROUND_NEAREST(_FP_I_TYPE wc, _FP_VAL X)
{
   if ((*_FP_FRAC_LOW(wc, X) & 15) != _FP_WORK_ROUND)
      _FP_FRAC_ADDI(wc, X, _FP_WORK_ROUND);
}

INLINE void _FP_ROUND_ZERO(_FP_I_TYPE wc, _FP_VAL X)
{
   /* do nothing */
}

INLINE void _FP_ROUND_PINF(_FP_I_TYPE wc, _FP_VAL X)
{
   if (!X->s && (*_FP_FRAC_LOW(wc,X) & 7))
      _FP_FRAC_ADDI(wc,X,_FP_WORK_LSB);
}

INLINE void _FP_ROUND_MINF(_FP_I_TYPE wc, _FP_VAL X)
{
   if (X->s && (*_FP_FRAC_LOW(wc,X) & 7))
      _FP_FRAC_ADDI(wc,X,_FP_WORK_LSB);
}

INLINE void _FP_ROUND(_FP_I_TYPE wc, _FP_VAL X)
{
   switch (FP_ROUNDMODE)
   {
      case FP_RND_NEAREST:
         _FP_ROUND_NEAREST(wc,X);
         break;
      case FP_RND_ZERO:
         _FP_ROUND_ZERO(wc,X);
         break;
      case FP_RND_PINF:
         _FP_ROUND_PINF(wc,X);
         break;
      case FP_RND_MINF:
         _FP_ROUND_MINF(wc,X);
         break;
   }
}

/* rest of this is from the old op-common.h */


/* Finish truly unpacking a native fp value by classifying the kind
 * of fp value and normalizing both the exponent and the fraction.
 */

INLINE void _FP_UNPACK_CANONICAL(char fs, _FP_I_TYPE wc, _FP_VAL X)
{
   /* We can't use a switch(X->e) here because _FP_EXPMAX(fs)
    * isn't a constant. [In fact it is, but the compiler can't
    * tell...]
    */
   if (X->e == 0)
   {
      if (_FP_FRAC_ZEROP(wc,X))
         X->c = FP_CLS_ZERO;
      else
      { /* a denormalised number */
         _FP_I_TYPE _shift = _FP_FRAC_CLZ(wc,X) - _FP_FRACXBITS(fs);
         _FP_FRAC_SLL(wc,X, _shift+_FP_WORKBITS);
         X->e -= _FP_EXPBIAS(fs) - 1 + _shift;
         X->c = FP_CLS_NORMAL;
      }
   }
   else if (X->e == _FP_EXPMAX(fs))
   {
      if (_FP_FRAC_ZEROP(wc,X))
         X->c = FP_CLS_INF; 
      else
         /* we don't differentiate between signaling and quiet NaNs */
         X->c = FP_CLS_NAN;
   }
   else
   { /* default: normal number */
      *_FP_FRAC_HIGH(wc,X) |= _FP_IMPLBIT(fs);
      _FP_FRAC_SLL(wc,X, _FP_WORKBITS);
      X->e -= _FP_EXPBIAS(fs);
      X->c = FP_CLS_NORMAL;
   }
}

/*
 * Before packing the bits back into the native fp result, take care
 * of such mundane things as rounding and overflow.  Also, for some
 * kinds of fp values, the original parts may not have been fully
 * extracted -- but that is ok, we can regenerate them now.
 */
INLINE void _FP_PACK_CANONICAL(char fs, _FP_I_TYPE wc, _FP_VAL X)
{
   switch(X->c)
   {
      case FP_CLS_NORMAL:
         X->e += _FP_EXPBIAS(fs);
         if (X->e > 0)
         {
            _FP_ROUND(wc,X);
            if (_FP_FRAC_OVERP(wc,fs,X))
            {
               _FP_FRAC_SRL(wc, X, _FP_WORKBITS+1);
               X->e++;
            }
            else
               _FP_FRAC_SRL(wc, X, _FP_WORKBITS);
            if (X->e >= _FP_EXPMAX(fs))
            {
               /* overflow to infinity */
               X->e = _FP_EXPMAX(fs);
               _FP_FRAC_SET_ZERO(wc, X);
            }
         }
         else
         { /* denormalized number */
            X->e = -X->e + 1;
            if (X->e <= _FP_WFRACBITS(fs))
            {
               _FP_FRAC_SRS(wc, X, X->e, _FP_WFRACBITS(fs));
               _FP_ROUND(wc,X);
               X->e = _FP_FRAC_OVERP(wc, fs, X);
               _FP_FRAC_SRL(wc, X, _FP_WORKBITS);
            }
            else
            {
               /* underflow to zero */
               X->e = 0;
               _FP_FRAC_SET_ZERO(wc, X);
            }
         }
         break;
         
      case FP_CLS_ZERO:
         X->e = 0;
         _FP_FRAC_SET_ZERO(wc, X);
         break;
         
      case FP_CLS_INF:
         X->e = _FP_EXPMAX(fs);
         _FP_FRAC_SET_ZERO(wc, X);
         break;
         
      case FP_CLS_NAN:
         X->e = _FP_EXPMAX(fs);
         if (!_FP_KEEPNANFRACP)
         {
            _FP_FRAC_SET_NAN(wc,fs,X);
            X->s = 0;
         }
         else
            *_FP_FRAC_HIGH(wc,X) |= _FP_QNANBIT(fs);
         break;
   }
}

/*
 * Main addition routine.  The input values should be cooked.
 */
INLINE void _FP_ADD(char fs, _FP_I_TYPE wc, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   switch (_FP_CLS_COMBINE(X->c, Y->c))
   {
      case _FP_CLS_COMBINE(FP_CLS_NORMAL, FP_CLS_NORMAL):
      {
         /* shift the smaller number so that its exponent matches the larger */
         _FP_I_TYPE diff = X->e - Y->e;
         if (diff < 0)
         {
            diff = -diff;
            if (diff <= _FP_WFRACBITS(fs))
               _FP_FRAC_SRS(wc,X,diff,_FP_WFRACBITS(fs));
            else if (!_FP_FRAC_ZEROP(wc,X))
               _FP_FRAC_SET_MIN(wc,X);
            else
               _FP_FRAC_SET_ZERO(wc,X);
            R->e = Y->e;
         }
         else
         {
            if (diff > 0)
            {
               if (diff <= _FP_WFRACBITS(fs))
                  _FP_FRAC_SRS(wc,Y,diff,_FP_WFRACBITS(fs));
               else if (!_FP_FRAC_ZEROP(wc,Y))
                  _FP_FRAC_SET_MIN(wc,Y);
               else
                  _FP_FRAC_SET_ZERO(wc,Y);
            }
            R->e = X->e;
         }
         
         R->c = FP_CLS_NORMAL;
         
         if(X->s == Y->s)
         {
            R->s = X->s;
            _FP_FRAC_ADD(wc,R,X,Y);
            if(_FP_FRAC_OVERP(wc,fs,R))
            {
               _FP_FRAC_SRS(wc, R, 1 ,_FP_WFRACBITS(fs));
               R->e++;
            }
         }
         else
         {
            R->s = X->s;
            _FP_FRAC_SUB(wc,R,X,Y);
            if (_FP_FRAC_ZEROP(wc,R))
            { /* return an exact zero */
               if (FP_ROUNDMODE == FP_RND_MINF)
                  R->s |= Y->s;
               else
                  R->s &= Y->s;
               R->c = FP_CLS_ZERO;
            }
            else
            {
               if (_FP_FRAC_NEGP(wc,R))
               {
                  _FP_FRAC_SUB(wc,R,Y,X);
                  R->s = Y->s;
               }
               
               /* renormalize after subtraction */
               diff = _FP_FRAC_CLZ(wc,R) - _FP_WFRACXBITS(fs);
               if (diff)
               {
                  R->e -= diff;
                  _FP_FRAC_SLL(wc,R,diff);
               }
            }
         }
         break;
      }
      
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_NAN):
         _FP_CHOOSENAN(fs,wc,R,X,Y);
         break;
         						
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_ZERO):	
         R->e = X->e;					
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_NORMAL):			
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_INF):			
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_ZERO):			
         _FP_FRAC_COPY(wc, R, X);					
         R->s = X->s;						
         R->c = X->c;						
         break;						
   
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_NORMAL):
         R->e = Y->e;
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_NAN):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_NAN):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_NAN):
         _FP_FRAC_COPY(wc, R, Y);
         R->s = Y->s;
         R->c = Y->c;
         break;

      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_INF):
         if (X->s != Y->s)
         { /* +INF + -INF => NaN */
            _FP_FRAC_SET_NAN(wc,fs,R);
            R->s = X->s ^ Y->s;
            R->c = FP_CLS_NAN;
            break;
         }
         /* fall through */

      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_NORMAL):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_ZERO):
         R->s = X->s;
         R->c = FP_CLS_INF;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_INF):
         R->s = Y->s;
         R->c = FP_CLS_INF;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_ZERO):
         /* make sure the sign is correct */
         if (FP_ROUNDMODE == FP_RND_MINF)
            R->s = X->s | Y->s;
         else
            R->s = X->s & Y->s;
         R->c = FP_CLS_ZERO;
         break;
         
      default:
         /* do nothing   abort(); */
   }
}

/*
 * Main negation routine.  FIXME -- when we care about setting exception
 * bits reliably, this will not do.  We should examine all of the fp classes.
 */
INLINE void _FP_NEG(char fs, _FP_I_TYPE wc, _FP_VAL R, _FP_VAL X)
{
   _FP_FRAC_COPY(wc, R, X);
   R->c = X->c;
   R->e = X->e;
   R->s = 1 ^ X->s;
}

/*
 * Main multiplication routine.  The input values should be cooked.
 */
INLINE void _FP_MUL(char fs, _FP_I_TYPE wc, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   R->s = X->s ^ Y->s;
   switch(_FP_CLS_COMBINE(X->c,Y->c))
   {
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_NORMAL):
         R->c = FP_CLS_NORMAL;
         R->e = X->e + Y->e + 1;
         
         _FP_MUL_MEAT(fs,R,X,Y);

         if (_FP_FRAC_OVERP(wc,fs,R))
            _FP_FRAC_SRS(wc, R, 1, _FP_WFRACBITS(fs));
         else
            R->e--;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_NAN):
         _FP_CHOOSENAN(fs, wc, R, X, Y);
         break;

      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_NORMAL):
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_ZERO):
         R->s = X->s;
         /* fall through */
         
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_NORMAL):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_NORMAL):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_ZERO):
         _FP_FRAC_COPY(wc,R,X);
         R->c = X->c;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_NAN):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_NAN):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_NAN):
         R->s = Y->s;
         /* fall through */

      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_ZERO):
         _FP_FRAC_COPY(wc,R,Y);
         R->c = Y->c;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_ZERO):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_INF):
         R->c = FP_CLS_NAN;
         _FP_FRAC_SET_NAN(wc,fs,R);
         break;
         
      default:
         /* do nothing   abort(); */
   }
}

/*
 * Main division routine.  The input values should be cooked.
 */
INLINE void _FP_DIV(char fs, _FP_I_TYPE wc, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   R->s = X->s + Y->s;
   switch(_FP_CLS_COMBINE(X->c, Y->c))
   {
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_NORMAL):
         R->c = FP_CLS_NORMAL;
         R->e = X->e - Y->e;
         _FP_DIV_MEAT(fs,R,X,Y);
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_NAN):
         _FP_CHOOSENAN(fs,wc,R,X,Y);
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_NORMAL):
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_NAN,FP_CLS_ZERO):
         R->s = X->s;
         _FP_FRAC_COPY(wc,R,X);
         R->c = X->c;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_NAN):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_NAN):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_NAN):
         R->s = Y->s;
         _FP_FRAC_COPY(wc,R,Y);
         R->c = Y->c;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_NORMAL):
         R->c = FP_CLS_ZERO;
         break;
		
      case _FP_CLS_COMBINE(FP_CLS_NORMAL,FP_CLS_ZERO):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_ZERO):
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_NORMAL):
         R->c = FP_CLS_INF;
         break;
         
      case _FP_CLS_COMBINE(FP_CLS_INF,FP_CLS_INF):
      case _FP_CLS_COMBINE(FP_CLS_ZERO,FP_CLS_ZERO):
         R->c = FP_CLS_NAN;
         _FP_FRAC_SET_NAN(wc,fs,R);
         break;
         
      default:
         /* can't happen! */
   }
}

/*
 * Main differential comparison routine.  The inputs should be raw not
 * cooked.  The return is -1,0,1 for normal values, whatever we're passed 
 * as un otherwise. [the reason for the un parameter is for efficiency:
 * it saves the caller having to convert the return code to the right thing
 */
INLINE _FP_I_TYPE _FP_CMP(char fs, _FP_I_TYPE wc, _FP_VAL X, _FP_VAL Y, _FP_I_TYPE un)
{
   /* NaNs are unordered... */
   if ((X->e == _FP_EXPMAX(fs) && !_FP_FRAC_ZEROP(wc,X))
      || (Y->e == _FP_EXPMAX(fs) && !_FP_FRAC_ZEROP(wc,Y)))
   {
      return un;
   }
   else
   {
      /* force -0 -> +0 : "Comparisons shall ignore the sign of zero" -- (5.7) */
      if (!X->e && _FP_FRAC_ZEROP(wc,X)) X->s = 0;
      if (!Y->e && _FP_FRAC_ZEROP(wc,Y)) Y->s = 0; /* was X->s = 0 in original function -- bug?? */
      
      if (X->s != Y->s)
         return X->s ? -1 : 1;
      else if (X->e > Y->e)
         return X->s ? -1 : 1;
      else if (X->e < Y->e)
         return X->s ? 1 : -1;
      else if (_FP_FRAC_GT(wc,X,Y))
         return X->s ? -1 : 1;
      else if (_FP_FRAC_GT(wc,Y,X))
         return X->s ? 1 : -1;
      else
         return 0;
   }
}
      
/* Simplification for strict equality.  */
INLINE _FP_I_TYPE _FP_CMP_EQ(char fs, _FP_I_TYPE wc, _FP_VAL X, _FP_VAL Y, _FP_I_TYPE un)
{
   /* NaNs are unordered */
   if ((X->e == _FP_EXPMAX(fs) && !_FP_FRAC_ZEROP(wc,X))
       || (Y->e == _FP_EXPMAX(fs) && !_FP_FRAC_ZEROP(wc,Y)))
   {
      return un;
   }
   else
   {
      return !(X->e == Y->e && _FP_FRAC_EQ(wc,X,Y)
              && (X->s == Y->s || (!X->e && _FP_FRAC_ZEROP(wc,X))));
   }
}

/*
 * Main square root routine.  The input value should be cooked.
 * 
 * All the exceptional cases are reasonably straightforward.
 * The positive normalised number algorithm is less so.
 * If X = mant * 2^{exp} where exp is even then 
 * R = sqrt(X) = sqrt(mant) * 2^{exp/2}
 * If exp is odd then R = sqrt(X) = sqrt(2*mant) * 2^{(exp-1)/2}
 * The code in _FP_SQRT() sets R's exponent to exp/2 or (exp-1)/2
 * as appropriate, and doubles the mantissa if necessary.
 * It then sets up for the _FP_SQRT_MEAT() routine by zeroing R,
 * setting q to the most significant bit of the mantissa
 * and shifting X left one place (to avoid overflow problems?)
 * The _FP_SQRT_MEAT() routine simply needs to do a fixed point
 * square root (with one more bit of precision than normal)
 * and return a properly normalised (including that extra bit)
 * result, adjusting the exponent again if needed.
 */
INLINE void _FP_SQRT(char fs, _FP_I_TYPE wc, _FP_VAL R, _FP_VAL X)
{
   _FP_W_TYPE q;
   
   switch(X->c)
   {
      case FP_CLS_NAN:
         R->s = 0;
         R->c = FP_CLS_NAN;
         _FP_FRAC_SET_ZERO(wc, X);
      case FP_CLS_INF:
         if (X->s)
         {
            R->s = 0;
            R->c = FP_CLS_NAN;                    /* sNAN */
         }
         else
         {
            R->s = 0;
            R->c = FP_CLS_INF;                    /* sqrt(+inf) = +inf */
         }
         break;
      case FP_CLS_ZERO:
         R->s = X->s;
         R->c = FP_CLS_ZERO;                      /* sqrt(+0) = +0, sqrt(-0) = -0 */
         break;
      case FP_CLS_NORMAL:
         R->s = 0;
         if (X->s)
         {
            R->c = FP_CLS_NAN;                    /* sNAN */
            break;
         }
         R->c = FP_CLS_NORMAL;
         if (X->e & 1)
            _FP_FRAC_SLL(wc,X,1);
         R->e = X->e >> 1;                        /* R->e = exp/2 or (exp-1)/2 */
         _FP_FRAC_SET_ZERO(wc,R);
         q = _FP_OVERFLOW(fs);
         _FP_FRAC_SLL(wc,X,1);
         _FP_SQRT_MEAT(wc,R,X,q);
         _FP_FRAC_SRL(wc,R,1);
   }
}
         
/*
 * Convert from FP to integer
 */

/* "When a NaN, infinity, large positive argument >= 2147483648.0, or 
 * large negative argument <= -2147483649.0 is converted to an integer,
 * the invalid_current bit...should be set and fp_exception_IEEE_754 should
 * be raised. If the floating point invalid trap is disabled, no trap occurs
 * and a numerical result is generated: if the sign bit of the operand
 * is 0, the result is 2147483647; if the sign bit of the operand is 1,
 * the result is -2147483648."
 * Similarly for conversion to extended ints, except that the boundaries
 * are >= 2^63, <= -(2^63 + 1), and the results are 2^63 + 1 for s=0 and
 * -2^63 for s=1.
 * -- SPARC Architecture Manual V9, Appendix B, which specifies how
 * SPARCs resolve implementation dependencies in the IEEE-754 spec.
 * I don't believe that the code below follows this. I'm not even sure
 * it's right! 
 * It doesn't cope with needing to convert to an n bit integer when there
 * is no n bit integer type. Fortunately gcc provides long long so this
 * isn't a problem for sparc32.
 * I have, however, fixed its NaN handling to conform as above.
 *         -- PMM 02/1998
 * NB: rsigned is not 'is r declared signed?' but 'should the value stored
 * in r be signed or unsigned?'. r is always(?) declared unsigned.
 * Comments below are mine, BTW -- PMM 
 * rsize is not the length of the actual type of the return value but
 * the length to which we convert; anything higher is (possibly) truncated.
 */
/* We convert to the _FP_CONV_I_TYPE; this might be long long if you
 * need conversion to 64bit int on a 32bit machine. SPARCV8 doesn't 
 * have the FDTOX/FXTOD &co insns so we can get away with making _FP_CONV_I_TYPE
 * 32bits in sfp-machine.h
 */
INLINE _FP_CONV_I_TYPE _FP_TO_INT(char fs, _FP_I_TYPE wc, _FP_VAL X, _FP_I_TYPE rsize, _FP_I_TYPE rsigned)
{
   _FP_CONV_I_TYPE r = 0;
   
   switch (X->c)
   {
      case FP_CLS_NORMAL:
         if (X->e < 0)
         {
            /* case FP_CLS_NAN: used to be here */
            case FP_CLS_ZERO:
               r = 0;
         }
         else if (X->e >= rsize - (rsigned != 0))
         {                                        /* overflow */
            case FP_CLS_NAN:
            case FP_CLS_INF:
               if (rsigned)
               {
                  r = 1;
                  r <<= rsize - 1;
                  r -= 1 - X->s;
               }
               else
               {
                  r = 0;
                  if (!X->s)
                     r = ~r;
               }
         }
         else
         {
            if (_FP_W_TYPE_SIZE * wc < rsize)
            {
               r = _FP_FRAC_ASSEMBLE(wc, X, rsize);
               r <<= X->e - _FP_WFRACBITS(fs);
            }
            else
            {
               if (X->e >= _FP_WFRACBITS(fs))
                  _FP_FRAC_SLL(wc,X,X->e - _FP_WFRACBITS(fs) + 1);
               else
                  _FP_FRAC_SRL(wc,X,_FP_WFRACBITS(fs) - X->e - 1);
               r = _FP_FRAC_ASSEMBLE(wc, X, rsize);
            }
            if (rsigned && X->s)
               r = -r;
         }
         break;
   }
   return r;
}

INLINE void _FP_FROM_INT(char fs, _FP_I_TYPE wc, _FP_VAL X, _FP_CONV_I_TYPE r, _FP_I_TYPE rsize)
{
   if (r)
   {
      X->c = FP_CLS_NORMAL;
      
      if ((X->s = (r<0)))
         r = -r;
      /* note that r is now considered unsigned, so we don't have to worry
       * about the single signed overflow case. */

      if (rsize <= _FP_W_TYPE_SIZE)
         X->e = __FP_CLZ(r);
      else
      {
         
#if _FP_CONV_I_TYPE_SIZE > _FP_W_TYPE_SIZE
         X->e = __FP_CLZ_2((_FP_W_TYPE)(r >> _FP_W_TYPE_SIZE), (_FP_W_TYPE) r);
#else  
         /* can't happen unless caller is buggy */
         printk("_FP_FROM_INT: bogus rsize (bug!)\n");
#endif    
      }
      
      if (rsize < _FP_W_TYPE_SIZE)
         X->e -= (_FP_W_TYPE_SIZE - rsize);
      X->e = rsize - X->e - 1;
      
      if (_FP_FRACBITS(fs) < rsize && _FP_WFRACBITS(fs) < X->e)
         r = __FP_FRAC_SRS_1(r,X->e - _FP_WFRACBITS(fs), rsize);
      r &= ~((_FP_W_TYPE)1 << X->e);
      _FP_FRAC_DISASSEMBLE(wc,X,(unsigned _FP_CONV_I_TYPE) r, rsize);
      _FP_FRAC_SLL(wc, X, _FP_WFRACBITS(fs) - X->e - 1);
   }
   else
   {
      X->e = FP_CLS_ZERO;
      X->s = 0;
   }
}

INLINE void FP_CONV(char dfs, char sfs, _FP_I_TYPE dwc, _FP_I_TYPE swc, _FP_VAL D, _FP_VAL S)
{
   _FP_FRAC_CONV(dwc,swc,dfs,sfs,D,S);
   D->e = S->e;
   D->c = S->c;
   D->s = S->s;
}
