/* op-dispatch-pre.h : dispatches calls to word-size or float type specific functions.
 * All calls to _FP_FOO_n() should be made from here and op-dispatch.h alone.
 * We rely on the wc parameter being constant and the 
 * compiler optimising away the if() and unused clauses.
 * Note that here we have to cast _FP_VAL to _FP_n. This
 * is perfectly safe; do not be alarmed :->
 * double.h & co should cast to _FP_VAL when calling generic routines,
 * and we cast to _FP_n here when calling specific routines. This
 * way, neither op-n.h nor op-common.h should have to have casts in them.
 *
 * The split between op-dispatch-pre.h and op-dispatch.h works like this:
 * + cpp macros (if any) have to go in -pre. 
 * + functions which are used by op-n.h have to go in -pre.
 * + functions which dispatch to other functions go in op-dispatch.h. 
 * Hopefully there won't be any circular dependencies :->
 */

INLINE _FP_W_TYPE _FP_IMPLBIT(char fs)
{
   if (fs == 's')
      return _FP_IMPLBIT_S;
   else if (fs == 'd')
      return _FP_IMPLBIT_D;
   else 
      return _FP_IMPLBIT_Q;
}

INLINE _FP_W_TYPE _FP_QNANBIT(char fs)
{
   if (fs == 's')
      return _FP_QNANBIT_S;
   else if (fs == 'd')
      return _FP_QNANBIT_D;
   else
      return _FP_QNANBIT_Q;
}

INLINE _FP_W_TYPE _FP_OVERFLOW(char fs)
{
   if (fs == 's')
      return _FP_OVERFLOW_S;
   else if (fs == 'd')
      return _FP_OVERFLOW_D;
   else 
      return _FP_OVERFLOW_Q;
}

INLINE _FP_I_TYPE _FP_FRACBITS(char fs)
{
   if (fs == 's')
      return _FP_FRACBITS_S;
   else if (fs == 'd')
      return _FP_FRACBITS_D;
   else 
      return _FP_FRACBITS_Q;
}

INLINE _FP_I_TYPE _FP_WFRACXBITS(char fs)
{
   if (fs == 's')
      return _FP_WFRACXBITS_S;
   else if (fs == 'd')
      return _FP_WFRACXBITS_D;
   else 
      return _FP_WFRACXBITS_Q;
}

INLINE _FP_I_TYPE _FP_WFRACBITS(char fs)
{
   if (fs == 's')
      return _FP_WFRACBITS_S;
   else if (fs == 'd')
      return _FP_WFRACBITS_D;
   else 
      return _FP_WFRACBITS_Q;
}

INLINE _FP_I_TYPE _FP_FRACXBITS(char fs)
{
   if (fs == 's')
      return _FP_FRACXBITS_S;
   else if (fs == 'd')
      return _FP_FRACXBITS_D;
   else 
      return _FP_FRACXBITS_Q;
}

INLINE _FP_I_TYPE _FP_EXPBIAS(char fs)
{
   if (fs == 's')
      return _FP_EXPBIAS_S;
   else if (fs == 'd')
      return _FP_EXPBIAS_D;
   else 
      return _FP_EXPBIAS_Q;
}

INLINE _FP_I_TYPE _FP_EXPMAX(char fs)
{
   if (fs == 's')
      return _FP_EXPMAX_S;
   else if (fs == 'd')
      return _FP_EXPMAX_D;
   else 
      return _FP_EXPMAX_Q;
}
