/* op-dispatch.h : dispatches calls to word-size or float type specific functions.
 * All calls to _FP_FOO_n() should be made from here and op-dispatch-pre.h alone.
 * We rely on the wc parameter being constant and the 
 * compiler optimising away the if() and unused clauses.
 * Note that here we have to cast _FP_VAL to _FP_n. This
 * is perfectly safe; do not be alarmed :->
 * double.h & co should cast to _FP_VAL when calling generic routines,
 * and we cast to _FP_n here when calling specific routines. This
 * way, neither op-n.h nor op-common.h should have to have casts in them.
 */

INLINE void _FP_FRAC_ADDI(_FP_I_TYPE wc, _FP_VAL X, _FP_I_TYPE I)
{
   if (wc == 1)
      _FP_FRAC_ADDI_1((_FP_1)X,I);
   else if (wc == 2)
      _FP_FRAC_ADDI_2((_FP_2)X,I);
   else
      _FP_FRAC_ADDI_4((_FP_4)X,I);
}

INLINE void _FP_FRAC_ADD(_FP_I_TYPE wc, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   if (wc == 1)
      _FP_FRAC_ADD_1((_FP_1)R, (_FP_1)X, (_FP_1)Y);
   else if (wc == 2)
      _FP_FRAC_ADD_2((_FP_2)R, (_FP_2)X, (_FP_2)Y);
   else
      _FP_FRAC_ADD_4((_FP_4)R, (_FP_4)X, (_FP_4)Y);
}

INLINE void _FP_FRAC_SUB(_FP_I_TYPE wc, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   if (wc == 1)
      _FP_FRAC_SUB_1((_FP_1)R, (_FP_1)X, (_FP_1)Y);
   else if (wc == 2)
      _FP_FRAC_SUB_2((_FP_2)R, (_FP_2)X, (_FP_2)Y);
   else
      _FP_FRAC_SUB_4((_FP_4)R, (_FP_4)X, (_FP_4)Y);
}

INLINE _FP_W_TYPE* _FP_FRAC_LOW(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      return _FP_FRAC_LOW_1((_FP_1)X);
   else if (wc == 2)
      return _FP_FRAC_LOW_2((_FP_2)X);
   else
      return _FP_FRAC_LOW_4((_FP_4)X);
}

INLINE _FP_W_TYPE* _FP_FRAC_HIGH(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      return _FP_FRAC_HIGH_1((_FP_1)X);
   else if (wc == 2)
      return _FP_FRAC_HIGH_2((_FP_2)X);
   else
      return _FP_FRAC_HIGH_4((_FP_4)X);
}

INLINE void _FP_FRAC_COPY(_FP_I_TYPE wc, _FP_VAL X, _FP_VAL Y)
{
   if (wc == 1)
      _FP_FRAC_COPY_1((_FP_1)X,(_FP_1)Y);
   else if (wc == 2)
      _FP_FRAC_COPY_2((_FP_2)X,(_FP_2)Y);
   else
      _FP_FRAC_COPY_4((_FP_4)X,(_FP_4)Y);
}

INLINE _FP_I_TYPE _FP_FRAC_CLZ(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      return _FP_FRAC_CLZ_1((_FP_1)X);
   else if (wc == 2)
      return _FP_FRAC_CLZ_2((_FP_2)X);
   else
      return _FP_FRAC_CLZ_4((_FP_4)X);
}

INLINE _FP_I_TYPE _FP_FRAC_ZEROP(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      return _FP_FRAC_ZEROP_1((_FP_1)X);
   else if (wc == 2)
      return _FP_FRAC_ZEROP_2((_FP_2)X);
   else
      return _FP_FRAC_ZEROP_4((_FP_4)X);
}

INLINE _FP_I_TYPE _FP_FRAC_NEGP(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      return _FP_FRAC_NEGP_1((_FP_1)X);
   else if (wc == 2)
      return _FP_FRAC_NEGP_2((_FP_2)X);
   else
      return _FP_FRAC_NEGP_4((_FP_4)X);
}

INLINE _FP_I_TYPE _FP_FRAC_OVERP(_FP_I_TYPE wc, char fs, _FP_VAL X)
{
   if (wc == 1)
      return _FP_FRAC_OVERP_1(fs, (_FP_1)X);
   else if (wc == 2)
      return _FP_FRAC_OVERP_2(fs, (_FP_2)X);
   else
      return _FP_FRAC_OVERP_4(fs, (_FP_4)X);
}

INLINE _FP_I_TYPE _FP_FRAC_GT(_FP_I_TYPE wc, _FP_VAL X, _FP_VAL Y)
{
   if (wc == 1)
      return _FP_FRAC_GT_1((_FP_1)X, (_FP_1)Y);
   else if (wc == 2)
      return _FP_FRAC_GT_2((_FP_2)X, (_FP_2)Y);
   else
      return _FP_FRAC_GT_4((_FP_4)X, (_FP_4)Y);
}

INLINE _FP_I_TYPE _FP_FRAC_GE(_FP_I_TYPE wc, _FP_VAL X, _FP_VAL Y)
{
   if (wc == 1)
      return _FP_FRAC_GE_1((_FP_1)X, (_FP_1)Y);
   else if (wc == 2)
      return _FP_FRAC_GE_2((_FP_2)X, (_FP_2)Y);
   else
      return _FP_FRAC_GE_4((_FP_4)X, (_FP_4)Y);
}

INLINE _FP_I_TYPE _FP_FRAC_EQ(_FP_I_TYPE wc, _FP_VAL X, _FP_VAL Y)
{
   if (wc == 1)
      return _FP_FRAC_EQ_1((_FP_1)X, (_FP_1)Y);
   else if (wc == 2)
      return _FP_FRAC_EQ_2((_FP_2)X, (_FP_2)Y);
   else
      return _FP_FRAC_EQ_4((_FP_4)X, (_FP_4)Y);
}

INLINE void _FP_FRAC_SLL(_FP_I_TYPE wc, _FP_VAL X, _FP_I_TYPE s)
{
   if (wc == 1)
      _FP_FRAC_SLL_1((_FP_1)X, s);
   else if (wc == 2)
      _FP_FRAC_SLL_2((_FP_2)X, s);
   else
      _FP_FRAC_SLL_4((_FP_4)X, s);
}

INLINE void _FP_FRAC_SRL(_FP_I_TYPE wc, _FP_VAL X, _FP_I_TYPE s)
{
   if (wc == 1)
      _FP_FRAC_SRL_1((_FP_1)X, s);
   else if (wc == 2)
      _FP_FRAC_SRL_2((_FP_2)X, s);
   else
      _FP_FRAC_SRL_4((_FP_4)X, s);
}

INLINE void _FP_FRAC_SRS(_FP_I_TYPE wc, _FP_VAL X, _FP_I_TYPE s, _FP_I_TYPE sz)
{
   if (wc == 1)
      _FP_FRAC_SRS_1((_FP_1)X, s, sz);
   else if (wc == 2)
      _FP_FRAC_SRS_2((_FP_2)X, s, sz);
   else
      _FP_FRAC_SRS_4((_FP_4)X, s, sz);
}

INLINE void _FP_FRAC_SET_ZERO(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      _FP_FRAC_SET_ZERO_1((_FP_1)X);
   else if (wc == 2)
      _FP_FRAC_SET_ZERO_2((_FP_2)X);
   else
      _FP_FRAC_SET_ZERO_4((_FP_4)X);
}

INLINE void _FP_FRAC_SET_NAN(_FP_I_TYPE wc, char fs, _FP_VAL X)
{
   if (wc == 1)
      _FP_FRAC_SET_NAN_1(fs, (_FP_1)X);
   else if (wc == 2)
      _FP_FRAC_SET_NAN_2(fs, (_FP_2)X);
   else
      _FP_FRAC_SET_NAN_4(fs, (_FP_4)X);
}

INLINE void _FP_FRAC_SET_MIN(_FP_I_TYPE wc, _FP_VAL X)
{
   if (wc == 1)
      _FP_FRAC_SET_MIN_1((_FP_1)X);
   else if (wc == 2)
      _FP_FRAC_SET_MIN_2((_FP_2)X);
   else
      _FP_FRAC_SET_MIN_4((_FP_4)X);
}

INLINE unsigned _FP_CONV_I_TYPE _FP_FRAC_ASSEMBLE(_FP_I_TYPE wc, _FP_VAL X, _FP_I_TYPE rsize)
{
    if (wc == 1)
      return _FP_FRAC_ASSEMBLE_1((_FP_1)X, rsize);
   else if (wc == 2)
      return _FP_FRAC_ASSEMBLE_2((_FP_2)X, rsize);
   else
      return _FP_FRAC_ASSEMBLE_4((_FP_4)X, rsize);
}

INLINE void _FP_FRAC_DISASSEMBLE(_FP_I_TYPE wc, _FP_VAL X, 
                                 unsigned _FP_CONV_I_TYPE r, _FP_I_TYPE rsize)
{
    if (wc == 1)
      _FP_FRAC_DISASSEMBLE_1((_FP_1)X, r, rsize);
   else if (wc == 2)
      _FP_FRAC_DISASSEMBLE_2((_FP_2)X, r, rsize);
   else
      _FP_FRAC_DISASSEMBLE_4((_FP_4)X, r, rsize);
}

INLINE void _FP_PACK_RAW(_FP_I_TYPE wc, char fs, void *val, _FP_VAL X)
{
   if (wc == 1)
      _FP_PACK_RAW_1(fs, val, (_FP_1)X);
   else if (wc == 2)
      _FP_PACK_RAW_2(fs, val, (_FP_2)X);
   else
      _FP_PACK_RAW_4(fs, val, (_FP_4)X);
}

INLINE void _FP_UNPACK_RAW(_FP_I_TYPE wc, char fs, _FP_VAL X, void* val)
{
   if (wc == 1)
      _FP_UNPACK_RAW_1(fs, (_FP_1)X, val);
   else if (wc == 2)
      _FP_UNPACK_RAW_2(fs, (_FP_2)X, val);
   else
      _FP_UNPACK_RAW_4(fs, (_FP_4)X, val);
}

INLINE void _FP_FRAC_CONV(_FP_I_TYPE dwc, _FP_I_TYPE swc, char dfs, char sfs,
                          _FP_VAL D, _FP_VAL S)
{
   switch (dwc | (swc << 3))
   {
      case 1 | (1<<3):
         _FP_FRAC_CONV_1_1(dfs, sfs, (_FP_1)D, (_FP_1)S);
         break;
       case 1 | (2<<3):
         _FP_FRAC_CONV_1_2(dfs, sfs, (_FP_1)D, (_FP_2)S);
         break;
      case 1 | (4<<3):
         _FP_FRAC_CONV_1_4(dfs, sfs, (_FP_1)D, (_FP_4)S);
         break;
      case 2 | (1<<3):
         _FP_FRAC_CONV_2_1(dfs, sfs, (_FP_2)D, (_FP_1)S);
         break;
      case 2 | (2<<3):
         _FP_FRAC_CONV_2_2(dfs, sfs, (_FP_2)D, (_FP_2)S);
         break;
      case 2 | (4<<3):
         _FP_FRAC_CONV_2_4(dfs, sfs, (_FP_2)D, (_FP_4)S);
         break;
      case 4 | (1<<3):
         _FP_FRAC_CONV_4_1(dfs, sfs, (_FP_4)D, (_FP_1)S);
         break;
      case 4 | (2<<3):
         _FP_FRAC_CONV_4_2(dfs, sfs, (_FP_4)D, (_FP_2)S);
         break;
      case 4 | (4<<3):
         _FP_FRAC_CONV_4_4(dfs, sfs, (_FP_4)D, (_FP_4)S);
         break;
      default:                                    /* can't happen */
         printk("_FP_FRAC_CONV: something is horribly wrong!\n");
         break;
   }
}

INLINE void _FP_SQRT_MEAT(_FP_I_TYPE wc, _FP_VAL R, _FP_VAL X, _FP_W_TYPE q)
{
   if (wc == 1)
      _FP_SQRT_MEAT_1((_FP_1)R, (_FP_1)X, q);
   else if (wc == 2)
      _FP_SQRT_MEAT_2((_FP_2)R, (_FP_2)X, q);
   else
      _FP_SQRT_MEAT_4((_FP_4)R, (_FP_4)X, q);
}

INLINE void _FP_MUL_MEAT(char fs, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   if (fs == 's')
      _FP_MUL_MEAT_S(R,X,Y);
   else if (fs == 'd')
      _FP_MUL_MEAT_D(R,X,Y);
   else
      _FP_MUL_MEAT_Q(R,X,Y);
}

INLINE void _FP_DIV_MEAT(char fs, _FP_VAL R, _FP_VAL X, _FP_VAL Y)
{
   if (fs == 's')
      _FP_DIV_MEAT_S(R,X,Y);
   else if (fs == 'd')
      _FP_DIV_MEAT_D(R,X,Y);
   else
      _FP_DIV_MEAT_Q(R,X,Y);
}
