/*
 * Definitions for IEEE Quad Precision
 */
#ifndef QUAD_H
#define QUAD_H

/* this one has to stay a macro, so the scope of the declaration is correct */
#define FP_DECL_Q(X)		_FP_DECL(__FP_Q_TYPE,X)

INLINE void FP_UNPACK_RAW_Q(__FP_Q_TYPE X, void *val)
{
   _FP_UNPACK_RAW(__FP_Q_SIZE,'q',(_FP_VAL)X,val);
}

INLINE void FP_PACK_RAW_Q(void *val, __FP_Q_TYPE X)
{
   _FP_PACK_RAW(__FP_Q_SIZE,'q',val,(_FP_VAL)X);
}


INLINE void FP_UNPACK_Q(__FP_Q_TYPE X, void *val)
{
   _FP_UNPACK_RAW(__FP_Q_SIZE,'q',(_FP_VAL)X,val);
   _FP_UNPACK_CANONICAL('q',__FP_Q_SIZE,(_FP_VAL)X);
}

INLINE void FP_PACK_Q(void *val, __FP_Q_TYPE X)
{
   _FP_PACK_CANONICAL('q',__FP_Q_SIZE,(_FP_VAL)X);
   _FP_PACK_RAW(__FP_Q_SIZE,'q',val,(_FP_VAL)X);
}

INLINE void FP_NEG_Q(__FP_Q_TYPE R, __FP_Q_TYPE X)
{
   _FP_NEG('q',__FP_Q_SIZE,(_FP_VAL)R, (_FP_VAL)X);
}

INLINE void FP_ADD_Q(__FP_Q_TYPE R, __FP_Q_TYPE X, __FP_Q_TYPE Y)
{
   _FP_ADD('q',__FP_Q_SIZE,(_FP_VAL)R, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_SUB_Q(__FP_Q_TYPE R, __FP_Q_TYPE X, __FP_Q_TYPE Y)
{
   /* _FP_SUB doesn't exist! 
    *   _FP_SUB('q',__FP_Q_SIZE,(_FP_VAL)R,(_FP_VAL)X,(_FP_VAL)Y);
    */
}

INLINE void FP_MUL_Q(__FP_Q_TYPE R, __FP_Q_TYPE X, __FP_Q_TYPE Y)
{
   _FP_MUL('q',__FP_Q_SIZE,(_FP_VAL)R, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_DIV_Q(__FP_Q_TYPE R, __FP_Q_TYPE X, __FP_Q_TYPE Y)
{
   _FP_DIV('q',__FP_Q_SIZE,(_FP_VAL)R, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_SQRT_Q(__FP_Q_TYPE R, __FP_Q_TYPE X)
{
   _FP_SQRT('q',__FP_Q_SIZE,(_FP_VAL)R, (_FP_VAL)X);
}

INLINE _FP_I_TYPE FP_CMP_Q(__FP_Q_TYPE X, __FP_Q_TYPE Y, _FP_I_TYPE un)
{
   return _FP_CMP('q', __FP_Q_SIZE, (_FP_VAL)X, (_FP_VAL)Y, un);
}

INLINE _FP_I_TYPE FP_CMP_EQ_Q(__FP_Q_TYPE X, __FP_Q_TYPE Y, _FP_I_TYPE un)
{
   return _FP_CMP_EQ('q', __FP_Q_SIZE, (_FP_VAL)X, (_FP_VAL)Y, un);
}

INLINE _FP_CONV_I_TYPE FP_TO_INT_Q(__FP_Q_TYPE X, _FP_I_TYPE rsz, _FP_I_TYPE rsigned)
{
   return _FP_TO_INT('q',__FP_Q_SIZE,(_FP_VAL)X,rsz,rsigned);
}

INLINE void FP_FROM_INT_Q(__FP_Q_TYPE X, _FP_CONV_I_TYPE r, _FP_I_TYPE rs)
{
   _FP_FROM_INT('q',__FP_Q_SIZE,(_FP_VAL)X,r,rs);
}

INLINE void FP_CONV_S_Q(__FP_S_TYPE D, __FP_Q_TYPE S)
{
   FP_CONV('s','q',__FP_S_SIZE,__FP_Q_SIZE,(_FP_VAL)D,(_FP_VAL)S);
}

INLINE void FP_CONV_D_Q(__FP_D_TYPE D, __FP_Q_TYPE S)
{
   FP_CONV('d','q',__FP_D_SIZE,__FP_Q_SIZE,(_FP_VAL)D,(_FP_VAL)S);
}

#endif /* not QUAD_H */
