/*
 * Definitions for IEEE Single Precision
 */
#ifndef SINGLE_H
#define SINGLE_H

#define FP_DECL_S(X)		_FP_DECL(__FP_S_TYPE,X)

INLINE void FP_UNPACK_RAW_S(__FP_S_TYPE X, void *val)
{
   _FP_UNPACK_RAW(__FP_S_SIZE,'s',(_FP_VAL)X,val);
}

INLINE void FP_PACK_RAW_S(void *val, __FP_S_TYPE X)
{
   _FP_PACK_RAW(__FP_S_SIZE,'s',val,(_FP_VAL)X);
}

INLINE void FP_UNPACK_S(__FP_S_TYPE X, void *val)
{
   _FP_UNPACK_RAW(__FP_S_SIZE,'s',(_FP_VAL)X,val);
   _FP_UNPACK_CANONICAL('s',__FP_S_SIZE,(_FP_VAL)X);
}

INLINE void FP_PACK_S(void *val, __FP_S_TYPE X)
{
   _FP_PACK_CANONICAL('s',__FP_S_SIZE,(_FP_VAL)X);
   _FP_PACK_RAW(__FP_S_SIZE,'s',val,(_FP_VAL)X);
}


INLINE void FP_NEG_S(__FP_S_TYPE R, __FP_S_TYPE X)
{
   _FP_NEG('s',__FP_S_SIZE,(_FP_VAL)R, (_FP_VAL)X);
}

INLINE void FP_ADD_S(__FP_S_TYPE R, __FP_S_TYPE X, __FP_S_TYPE Y)
{
   _FP_ADD('s',__FP_S_SIZE,(_FP_VAL)R, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_SUB_S(__FP_S_TYPE R, __FP_S_TYPE X, __FP_S_TYPE Y)
{
   /* _FP_SUB doesn't exist! 
    *   _FP_SUB('s',__FP_S_SIZE,(_FP_VAL)R,(_FP_VAL)X,(_FP_VAL)Y);
    */
}

INLINE void FP_MUL_S(__FP_S_TYPE R, __FP_S_TYPE X, __FP_S_TYPE Y)
{
   _FP_MUL('s',__FP_S_SIZE,(_FP_VAL)R, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_DIV_S(__FP_S_TYPE R, __FP_S_TYPE X, __FP_S_TYPE Y)
{
   _FP_DIV('s',__FP_S_SIZE,(_FP_VAL)R, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_SQRT_S(__FP_S_TYPE R, __FP_S_TYPE X)
{
   _FP_SQRT('s',__FP_S_SIZE,(_FP_VAL)R, (_FP_VAL)X);
}

INLINE _FP_I_TYPE FP_CMP_S(__FP_S_TYPE X, __FP_S_TYPE Y, _FP_I_TYPE un)
{
   return _FP_CMP('s',__FP_S_SIZE, (_FP_VAL)X, (_FP_VAL)Y, un);
}

INLINE _FP_I_TYPE FP_CMP_EQ_S(__FP_S_TYPE X, __FP_S_TYPE Y, _FP_I_TYPE un)
{
   return _FP_CMP_EQ('s', __FP_S_SIZE, (_FP_VAL)X, (_FP_VAL)Y, un);
}

INLINE _FP_CONV_I_TYPE FP_TO_INT_S(__FP_S_TYPE X, _FP_I_TYPE rsz, _FP_I_TYPE rsigned)
{
   return _FP_TO_INT('s',__FP_S_SIZE,(_FP_VAL)X,rsz,rsigned);
}

INLINE void FP_FROM_INT_S(__FP_S_TYPE X, _FP_CONV_I_TYPE r, _FP_I_TYPE rs)
{
   _FP_FROM_INT('s',2,(_FP_VAL)X,r,rs);
}

INLINE void FP_CONV_D_S(__FP_D_TYPE X, __FP_S_TYPE Y)
{
   FP_CONV('d','s', __FP_D_SIZE, __FP_S_SIZE, (_FP_VAL)X, (_FP_VAL)Y);
}

INLINE void FP_CONV_Q_S(__FP_Q_TYPE X, __FP_S_TYPE Y)
{
   FP_CONV('q','s', __FP_Q_SIZE, __FP_S_SIZE, (_FP_VAL)X, (_FP_VAL)Y);
}


#endif /* not SINGLE_H */
