/* $Id: ptifddi.c,v 1.1.2.4 1997/03/03 07:00:50 davem Exp $
 * ptifddi.c: Network driver for Performance Technologies single-attach
 *            and dual-attach FDDI sbus cards.
 *
 * Copyright (C) 1996 David S. Miller (davem@caip.rutgers.edu)
 */

static char *version =
        "ptifddi.c:v1.0 10/Dec/96 David S. Miller (davem@caipfs.rutgers.edu)\n";

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>
#include <asm/byteorder.h>
#include <asm/checksum.h>

#include <asm/idprom.h>
#include <asm/sbus.h>
#include <asm/openprom.h>
#include <asm/oplib.h>
#include <asm/auxio.h>
#include <asm/system.h>
#include <asm/pgtable.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include "ptifddi.h"

#include "ptifddi_asm.h"

#define DEBUG_DETECT
#define DEBUG_TEST
/* #define DEBUG_IRQ */
/* #define DEBUG_TRANSMIT */
/* #define DEBUG_RECEIVE */
/* #define DEBUG_HEADER */

#ifdef DEBUG_DETECT
#define DET(x)   printk x
#else
#define DET(x)
#endif

#ifdef DEBUG_TEST
#define DTST(x)   printk x
#else
#define DTST(x)
#endif

#ifdef DEBUG_IRQ
#define DIRQ(x)  printk x
#else
#define DIRQ(x)
#endif

#ifdef DEBUG_TRANSMIT
#define DTX(x)  printk x
#else
#define DTX(x)
#endif

#ifdef DEBUG_RECEIVE
#define DRX(x)  printk x
#else
#define DRX(x)
#endif

#ifdef DEBUG_HEADER
#define DHDR(x) printk x
#else
#define DHDR(x)
#endif

#ifdef MODULE
static struct ptifddi *root_pti_dev = NULL;
#endif

static void dump_dpram(struct ptifddi *pp)
{
	struct dfddi_ram *rp = pp->dpram;

	printk("\n");
	printk("DPRAM: regs[%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x]\n",
	       rp->regs[0], rp->regs[1], rp->regs[2], rp->regs[3],
	       rp->regs[4], rp->regs[5], rp->regs[6], rp->regs[7]);
	printk("DPRAM: re_sync[%02x] tswitch[%02x] evq_lost[%04x] signal_lost[%04x]\n",
	       rp->re_sync, rp->tswitch, rp->evq_lost, rp->signal_lost);
	printk("DPRAM: lerror[%04x] rstate[%04x] dswitch[%04x] cbusy[%02x] hbusy[%02x]\n",
	       rp->lerror, rp->rstate, rp->dswitch, rp->cbusy, rp->hbusy);
	printk("DPRAM: istat[%04x] disable[%04x] ucvalid[%04x] u0addr[%08x]\n",
	       rp->istat, rp->disable, rp->ucvalid, rp->u0addr);
	printk("DPRAM: timerint[%04x,%04x] P_player[%08x] Q_player[%08x] maxint[%04x,%04x]\n",
	       rp->timerint[0], rp->timerint[1], rp->P_player, rp->Q_player,
	       rp->maxint[0], rp->maxint[1]);
	printk("DPRAM: reset[%04x] iack[%04x] loader_addr[%04x]\n",
	       rp->reset, rp->iack, rp->loader_addr);
}


static inline void pti_reset(struct ptifddi *pp)
{
	*pp->reset = 1;
}

static inline void pti_unreset(struct ptifddi *pp)
{
	*pp->unreset = 1;
}

static inline unsigned short pti_get_iack(struct ptifddi *pp)
{
	unsigned short result;

	result = pp->dpram->iack;

	return result;
}

static void pti_load_code_base(struct dfddi_ram *rp, unsigned short addr)
{
	rp->loader_addr = addr;
}

#define REQUEST_TIMEOUT 1000000

static void pti_do_request(struct ptifddi *pp, unsigned char wait, unsigned int cmd)
{
	struct dfddi_ram *rp = pp->dpram;
	unsigned short aa, bb, cc;
	unsigned char aaa, bbb, ccc, ddd;
	volatile unsigned char *u0bytes = (volatile unsigned char *) &rp->u0addr;
	volatile unsigned short *ishorts;
	inst command;
	unsigned long flags;
	int i;

	ishorts = (volatile unsigned short *) &rp->instructions[0];
	command.word = cmd;

	save_and_cli(flags);
	for(i = 0; i < REQUEST_TIMEOUT; i++) {
		aaa = u0bytes[0];
		bbb = u0bytes[3];
		ddd = u0bytes[2];
		ccc = bbb & 0x80;
		aaa = (aaa ^ ddd);
		if(!(aaa & (!ccc)))
			break;
	}
	if(i == REQUEST_TIMEOUT) {
		restore_flags(flags);
		return;
	}

	ishorts[1] = command.shorts[0];
	ishorts[0] = command.shorts[1];

	u0bytes[3] |= 0x80;
	u0bytes[0] = ((0x200 >> 2) & 0xff);
	u0bytes[2] = (((0x200 + 4) >> 2) & 0xff);
	u0bytes[3] = 0x3e;

	if(wait) {
		for(i = 0; i < REQUEST_TIMEOUT; i++) {
			aaa = u0bytes[0];
			bbb = u0bytes[3];
			ddd = u0bytes[2];
			ccc = bbb & 0x80;
			aaa = (aaa ^ ddd);
			if(!(aaa && (!ccc)))
				break;
		}
		if(i = REQUEST_TIMEOUT) {
			restore_flags(flags);
			return;
		}
	}
	restore_flags(flags);
}

static unsigned short pti_answer(struct ptifddi *pp, unsigned short where)
{
	unsigned char c;

	pp->maddr.b = ((unsigned char *)pp->dpram) + where;
	pp->rq.answer.bb[0] = *(pp->maddr.b + 1);
	pp->rq.answer.bb[1] = *(pp->maddr.b);

	return pp->rq.answer.ww[0];
}

static inline void Reqmovr(struct ptifddi *pp, unsigned short a, unsigned short b)
{
	unsigned int cmd = 0x78000000;

	cmd |= (((a & 0x7ff) << 16) | 0x800 | (b & 0x7ff));
	pti_do_request(pp, 1, cmd);
}

static inline void Reqxchr(struct ptifddi *pp, unsigned short a, unsigned short b)
{
	unsigned int cmd = 0x78000000;

	cmd |= (((a & 0x7ff)<<16) | 0x1000 | (b & 0x7ff));
	pti_do_request(pp, 1, cmd);
}

static inline void Reqxorr(struct ptifddi *pp, unsigned short a, unsigned short b)
{
	unsigned int cmd = 0x78000000;

	cmd |= (((a & 0x7ff)<<16) | 0x7000 | (b & 0x7ff));
	pti_do_request(pp, 1, cmd);
}

static inline void Reqmovc(struct ptifddi *pp, unsigned short a, unsigned short b)
{
	unsigned int cmd = 0x08000000;

	cmd |= (((a & 0x7ff)<<16) | (b & 0xffff));
	pti_do_request(pp, 1, cmd);
}

static inline void Reqxchc(struct ptifddi *pp, unsigned short a, unsigned short b)
{
	unsigned int cmd = 0x10000000;

	cmd |= (((a & 0x7ff)<<16) | (b & 0xffff));
	pti_do_request(pp, 1, cmd);
}

static inline void Reqxorc(struct ptifddi *pp, unsigned short a, unsigned short b)
{
	unsigned int cmd = 0x70000000;

	cmd |= (((a & 0x7ff)<<16) | (b & 0xffff));
}

/* Request attribute modifiers. */
#define refP(val)	 (val)
#define refQ(val)	((val) + 0x040)
#define refM(val)	((val) + 0x100)
#define refB(val)	((val) + 0x1c0)
#define refD(val)	((val) + 0x400)
#define refS(val)	((val) + 0x500)

static inline void pti_clear_dpram(struct ptifddi *pp)
{
	memset(pp->dpram, 0, DPRAM_SIZE);
}

/* The card test takes a very long time... */
#define CARD_TEST_TIMEOUT	100000000

static int pti_card_test(struct ptifddi *pp)
{
	struct dfddi_ram *rp		= pp->dpram;
	volatile unsigned char *code	= &rp->instructions[0];
	volatile unsigned char *status	= (volatile unsigned char *) rp;
	unsigned char last		= 0;
	int i, result;
	int clicks			= CARD_TEST_TIMEOUT;

	pti_reset(pp);
	pti_clear_dpram(pp);
	for(i = 0; i < sizeof(test_firmware); i++)
		code[i] = test_firmware[i];
	pti_load_code_base(rp, test_firmware_dev_addr);

	clicks = CARD_TEST_TIMEOUT;
	*status = 0;

	/* Reset single attach state machine before the test. */
	rp->reset = DPRAM_RESET_SINGLE;

	/* Unreset, to get the test code running. */
	pti_unreset(pp);

	/* Wait for dpram status to become 5, else fail if we time out. */
#ifdef DEBUG_TEST
	dump_dpram(pp);
#endif
	result = 1;
	while(--clicks) {
		unsigned char sval = *status;

		if(sval == 5) {
#ifdef DEBUG_TEST
			dump_dpram(pp);
#endif
			result = 0;
			break;
		}
		if((sval != 0) && (sval != last)) {
			DTST(("sval[%02x]", sval));

			/* This bit set means error during testing. */
			if(sval & 0x40)
				break;
		}
		last = sval;
	}
#ifdef DEBUG_TEST
	dump_dpram(pp);
#endif
	pti_reset(pp);
	return result;
}

static void pti_init_firmware_loader(struct ptifddi *pp)
{
	struct dfddi_ram *rp = pp->dpram;
	int i;

	for(i = 0; i < sizeof(firmware_loader); i++)
		rp->loader.loader_firmware[i] = firmware_loader[i];
	pti_load_code_base(rp, firmware_loader_dev_addr);
}

static void pti_load_main_firmware(struct ptifddi *pp)
{
	struct dfddi_ram *rp		= pp->dpram;
	struct dpram_loader *lp		= &rp->loader;
	int i;

	lp->dpram_stat = DPRAM_STAT_EMPTY;
	pti_unreset(pp);

	printk("\npti_load_main_firmware: [");
	for(i = 0; i < main_firmware_size; i += 16) {
		/* Don't overwrite the firmware loader. */
		if(((main_firmware_dev_addr + i) >= 0xf774) &&
		   ((main_firmware_dev_addr + i) < 0xffd0))
			continue;

		while(lp->dpram_stat != DPRAM_STAT_EMPTY)
			barrier();

		lp->addr_low = (((main_firmware_dev_addr + i) >> 8) & 0xff);
		lp->addr_hi  =  ((main_firmware_dev_addr + i) & 0xff);
		lp->num_bytes = 16;
		lp->data[0x0] = main_firmware[i + 0x0];
		lp->data[0x1] = main_firmware[i + 0x1];
		lp->data[0x2] = main_firmware[i + 0x2];
		lp->data[0x3] = main_firmware[i + 0x3];
		lp->data[0x4] = main_firmware[i + 0x4];
		lp->data[0x5] = main_firmware[i + 0x5];
		lp->data[0x6] = main_firmware[i + 0x6];
		lp->data[0x7] = main_firmware[i + 0x7];
		lp->data[0x8] = main_firmware[i + 0x8];
		lp->data[0x9] = main_firmware[i + 0x9];
		lp->data[0xa] = main_firmware[i + 0xa];
		lp->data[0xb] = main_firmware[i + 0xb];
		lp->data[0xc] = main_firmware[i + 0xc];
		lp->data[0xd] = main_firmware[i + 0xd];
		lp->data[0xe] = main_firmware[i + 0xe];
		lp->data[0xf] = main_firmware[i + 0xf];

		lp->dpram_stat = DPRAM_STAT_VALID;
		printk(".");
	}

	printk("] ");

	/* Wait for the last data block to be transferred. */
	while(lp->dpram_stat != DPRAM_STAT_EMPTY)
		barrier();

	/* Place card back into reset state. */
	pti_reset(pp);

	printk("done\n");
}

static void pti_load_smt_firmware(struct ptifddi *pp)
{
	struct dfddi_ram *rp		= pp->dpram;
	struct dpram_loader *lp		= &rp->loader;
	int i;

	lp->dpram_stat = DPRAM_STAT_EMPTY;

	for(i = 0; i < smt_firmware_size; i += 16) {
		while(lp->dpram_stat != DPRAM_STAT_EMPTY)
			udelay(10);

		lp->addr_low = ((smt_firmware_dev_addr + i) >> 8);
		lp->addr_hi = ((smt_firmware_dev_addr + i) & 0xff);
		lp->smt_nbytes = 16;
		lp->data[0x0] = smt_firmware[i + 0x0];
		lp->data[0x1] = smt_firmware[i + 0x1];
		lp->data[0x2] = smt_firmware[i + 0x2];
		lp->data[0x3] = smt_firmware[i + 0x3];
		lp->data[0x4] = smt_firmware[i + 0x4];
		lp->data[0x5] = smt_firmware[i + 0x5];
		lp->data[0x6] = smt_firmware[i + 0x6];
		lp->data[0x7] = smt_firmware[i + 0x7];
		lp->data[0x8] = smt_firmware[i + 0x8];
		lp->data[0x9] = smt_firmware[i + 0x9];
		lp->data[0xa] = smt_firmware[i + 0xa];
		lp->data[0xb] = smt_firmware[i + 0xb];
		lp->data[0xc] = smt_firmware[i + 0xc];
		lp->data[0xd] = smt_firmware[i + 0xd];
		lp->data[0xe] = smt_firmware[i + 0xe];
		lp->data[0xf] = smt_firmware[i + 0xf];

		lp->dpram_stat = DPRAM_STAT_VALID;

		/* Tell the board to load this SMT firmware block. */
		pti_do_request(pp, 1, (INST_CALL | CODE_ASM | (HLOADER >> 2)));
	}
}

static void pti_set_fware_irq_handlers(struct ptifddi *pp)
{
	struct dfddi_ram *dp = pp->dpram;

	dp->maxint[1] = 0x8000;
	dp->maxint[0] = (CODE_ASM | (0xf708 >> 2));
	dp->maxint[1] = 0x0000;

	dp->timerint[1] = 0x8000;
	dp->timerint[0] = (CODE_ASM | (0xf700 >> 2));
	dp->timerint[1] = 0x0000;
}

static inline int pti_remove(struct ptifddi *pp, int which)
{
	unsigned long flags;
	int ret = 0;

	save_flags(flags); cli();
	if(pp->size[which]) {
		if(!--pp->size[which])
			pp->last[which] = 0;

		ret = pp->first[which];

		pp->first[which] = pp->dirtry[ret];
		pp->dirtry[ret] = ret;
	}
	restore_flags(flags);
	return ret;
}

static inline void pti_append(struct ptifddi *pp, int where, int buf)
{
	unsigned long flags;

	save_flags(flags); cli();
#ifdef DEBUG_CHAINS
	if((where > NUMCHAINS) || !buf || (buf > pp->dirtry[0]) ||
	   (pp->dirtry[buf] != buf)) {
		printk("ptifddi: Chain bug [w(%d),b(%d)] d[0](%d) d[b](%d)\n",
		       where, buf, pp->dirtry[0], pp->dirtry[buf]);
		restore_flags(flags);
		return;
	}
#endif
	if(!pp->size[where]++)
		pp->first[where] = buf;
	else
		pp->dirtry[pp->last[where]] = buf;

	pp->dirtry[buf] = 0;
	pp->last[where] = buf;
	restore_flags(flags);
}

static inline void pti_init_command_queue(struct ptifddi *pp)
{
	struct command_queue *que, *qua;
	int i, b;

	b = MAX_COMMQUEUE;
	que = pp->command_queue;
	pp->cdqwr = que;
	pp->cdqrd = que;
	qua = que;

	for(i = 0; i < (b - 1); i++) {
		qua->next = ++que;
		qua++;
	}
	que->next = pdrps->cdqwr;
}

static inline void pti_init_buffpool(struct ptifddi *pp, entry psize)
{
	entry h;

	if(psize == 0 || psize > POOLSPACE)
		psize = POOLSPACE;

	pp->dirtry[0] = psize;
	for(h = 1; h < psize; h++)
		pp->dirtry[h] = (h + 1);
	pp->dirtry[psize] = 0;

	for(h = 1; h <= NUMCHAINS; h++) {
		pp->size[h]	= 0;
		pp->last[h]	= 0;
		pp->first[h]	= 0;
	}
	pp->size[0]	= psize;
	pp->last[0]	= psize;
	pp->first[0]	= 1;
}

static int pti_allocate(struct ptifddi *pp)
{
	unsigned int *dma_mem;
	unsigned char *area;

	dma_mem = sparc_dvma_malloc(0x1000 * AREASIZE, "PTI FDDI Buffers");
	if(!dma_mem)
		return 1;

	pp->areapoi = dma_mem;
	pp->area    = pp->areapoi;

	area = pp->area;
	pp->mlbx.l = (unsigned int *)area;
	pp->command_queue = (struct command_queue *) (area + sizeof(pp->mlbx.l));
	pti_init_command_queue(pp);

	area += 0x1000;

	for(i = 0; i < 2; i++) {
		pp->trch[i].odudb	= (struct odud *) area;
		pp->trch[i].odud_rd	= (struct odud *) area;
		pp->trch[i].odud_wr	= (struct odud *) area;
		area += 0x1000;

		pp->trch[i].odude	= (struct odud *) area;
		pp->trch[i].cnfb	= (struct cnf *)  area;
		pp->trch[i].cnf_wr	= (struct cnf *)  area;
		area += 0x1000;

		pp->trch[i].cnfe	= (struct cnf *)  area;
		pp->trch[i].cnfl	= 0x1fd;
		pp->trch[i].cnfn	= 0x1fd;
		pp->trch[i].reqb	= (struct req *)  area;
		pp->trch[i].req_rd	= (struct req *)  area;
		pp->trch[i].req_wr	= (struct req *)  area;
		area += 0x1000;

		pp->trch[i].reqe	= (struct req *)  area;
		pp->trch[i].reql	= 0x1ff;
	}

	for(i = 0; i < 3; i++) {
		pp->rcvch[i].idudb	= (struct idud *) area;
		pp->rcvch[i].idud_rd	= (struct idud *) area;
		area += 0x1000;

		pp->rcvch[i].idude	= (struct idud *) area;
		pp->rcvch[i].idudl	= 0x1fd;
		pp->rcvch[i].pspb	= (struct psp *)  area;
		pp->rcvch[i].psp_rd	= (struct psp *)  area;
		area += 0x1000;

		pp->rcvch[i].pspe	= (struct psp *)  area;
		pp->rcvch[i].pspl	= 0xffff;
		pp->rcvch[i].pspn	= 0;
	}

	pp->pstart   = area;
	pp->buffsize = BPSIZE;

	pti_init_buffpool(pp, POOLSPACE);

	return 0;
}

static unsigned char *pti_get_rx_buff(struct ptifddi *pp)
{
	int entry = pti_remove(pp, RX_CHAIN);
	if(entry)
		return ((unsigned char *)(pp->poolstart + ((entry - 1) * pp->buffsize)));
	else
		return (unsigned char *) 0;
}

static unsigned char *pti_get_tx_buff(struct ptifddi *pp)
{
	int entry = pti_remove(pp, TX_CHAIN);
	if(entry)
		return ((unsigned char *)(pp->poolstart + ((entry - 1) * pp->buffsize)));
	else
		return (unsigned char *) 0;
}

static void pti_set_channel(struct ptifddi *pp)
{
	union addr dma_mbox;
	unsigned short response;

	if(pp->fddi_smt && !pp->fatalerr) {
		Reqxchc(pp, refP(1), 0xe1);
		Reqxchc(pp, refP(3), 0);
		Reqxchc(pp, refP(4), 0);

		pp->recenb = 1;

		if(pp->optwis == 1) {
			Reqmovr(refD(10), refP(0x3c));
			response = pti_answer(pp, 20);
			Reqxchc(refP(0x3c), (response | 8) & 0xff);
			Reqmovr(refD(10), refP(0x3b));
			response = pti_answer(pp, 20);
			Reqxchc(refP(0x3b), (response | 8) & 0xff);
		}

		Reqxchc(refP(0), 0x21);

		if(dual == 2) {
			Reqxchc(refQ(1), 0);
			Reqxchc(refQ(3), 0);
			Reqxchc(refQ(4), 0);

			if(pp->optwis == 1) {
				Reqmovr(refD(10), refQ(0x3c));
				response = pti_answer(pp, 20);
				Reqxchc(refQ(0x3c), (response | 8) & 0xff);
				Reqmovr(refD(10), refQ(0x3b));
				response = pti_answer(pp, 20);
				Reqxchc(refQ(0x3b), (response | 8) & 0xff);
			}

			Reqxchc(refQ(0), 0x21);
		}
		pti_do_request(pp, 0, CMCNT);
		for(i = 0; i < 6; i++) {
			Reqxchc(refM(0x40 + i), pp->own_address[i]);
		}
		pti_do_request(pp, 0, GADDR);
		Reqxchc(refM(1), 2);
		Reqxchc(refM(0x11), 3);
	}

	Reqxchc(refB(2), 0);
	if(pp->bursts & DMA_BURST32)
		Reqxchc(refB(0), 0x10);
	else
		Reqxchc(refB(0), 0x90);

	Reqxchc(refB(1), 0xff);
	Reqxchc(refB(5), 0);
	Reqxchc(refB(0x1a), 0x14);
	Reqxchc(refB(0x17), 0x19);
	Reqxchc(refB(0x1b), 0);
	Reqxchc(refB(0x1c), 0);
	Reqxchc(refB(0x18), 0x41);
	Reqxchc(refB(0x10), 0);
	Reqxchc(refB(0x11), 0x20);
	Reqxchc(refB(0x12), 0xc0);
	Reqxchc(refB(0x13), 0xc0);
	Reqxchc(refB(0x16), 1);
	Reqxchc(refB(7), 0xff);
	Reqxchc(refB(9), 0);
	Reqxchc(refB(0xb), 0xd5);
	Reqxchc(refB(0xf), 0xaa);
	Reqxchc(refB(0x15), 0x3b);

	dma_mbox.b = (unsigned char *) pp->mlbx.b;
	*((unsigned short *)&pp->dpram.regs[6]) = (((unsigned int)dma_mbox.b)>>16);
	*((unsigned short *)&pp->dpram.regs[4]) = ((unsigned int)dma_mbox.b);
	pti_do_request(pp, 1, MLBX+2);

	*(pp->mlbx.l) = ((unsigned int)pp->trch[0].cnfb);
	pti_do_request(pp, 1, PNTER+6);
	*(pp->mlbx.l) = ((unsigned int)pp->trch[0].reqb);
	pti_do_request(pp, 1, PNTER+7);
	*(pp->mlbx.l) = ((unsigned int)pp->trch[1].cnfb);
	pti_do_request(pp, 1, PNTER+2);
	*(pp->mlbx.l) = ((unsigned int)pp->trch[1].reqb);
	pti_do_request(pp, 1, PNTER+3);
	*(pp->mlbx.l) = ((unsigned int)pp->rcvch[0].idudb);
	pti_do_request(pp, 1, PNTER+0x11);
	*(pp->mlbx.l) = ((unsigned int)pp->rcvch[0].pspb);
	pti_do_request(pp, 1, PNTER+0x12);

	*(pp->mlbx.l) = ((unsigned int)pp->rcvch[0].idudb);
	pti_do_request(pp, 1, PNTER+0xd);
	*(pp->mlbx.l) = ((unsigned int)pp->rcvch[0].pspb);
	pti_do_request(pp, 1, PNTER+0xe);
	*(pp->mlbx.l) = ((unsigned int)pp->rcvch[2].idudb);
	pti_do_request(pp, 1, PNTER+9);
	*(pp->mlbx.l) = ((unsigned int)pp->rcvch[2].pspb);
	pti_do_request(pp, 1, PNTER+0xa);

	pti_do_request(pp, 1, LIMIT | (3 << 16) | (pp->trch[0].cnfl & 0x1ff));
	pti_do_request(pp, 1, LIMIT | (2 << 16));
	pti_do_request(pp, 1, LIMIT | (1 << 16) | (pp->trch[1].cnfl & 0x1ff));
	pti_do_request(pp, 1, LIMIT);

	pti_do_request(pp, 1, LIMIT | (4 << 16) | (pp->rcvch[2].idudl & 0x1ff));
	pti_do_request(pp, 1, LIMIT | (5 << 16) | (pp->rcvch[2].pspl & 0x1ff));

	pti_do_request(pp, 1, LIMIT | (6 << 16) | (pp->rcvch[1].idudl & 0x1ff));
	pti_do_request(pp, 1, LIMIT | (7 << 16) | (pp->rcvch[1].pspl & 0x1ff));
	pti_do_request(pp, 1, LIMIT | (8 << 16) | (pp->rcvch[0].idudl & 0x1ff));
	pti_do_request(pp, 1, LIMIT | (9 << 16) | (pp->rcvch[0].pspl & 0x1ff));

	Reqmovr(pp, refD(10), refB(0xa));
	response = (pti_answer(pp, 20) & ~(0xd5));
	Reqxchc(pp, refB(0xa), a & 0xff);

	Reqmovr(pp, refD(10), refB(0xe));
	response = (pti_answer(pp, 20) & ~(0x88));
	Reqxchc(pp, refB(0xe), a & 0xff);
	Reqxchc(pp, refB(6), 0);
	Reqxchc(pp, refB(5), 0xd8);

	if(!pp->fddi_smt && !pp->fatalerr) {
		pp->fddi_noring = 1;
		Reqxchc(pp, refM(0x2f), 1);
		Reqxchc(pp, refM(0), 1);
	}
}

static void pti_fddi_go(struct ptifddi *pp)
{
	struct dfddi_ram *dp = pp->dpram;
	unsigned short response;
	struct psp *rxd;
	int i;

	Reqxchc(pp, refB(0), 4);
	Reqmovr(pp, RefD(4), RefB(0));
	response = pti_get_response(pp, 8);
	while(response & 4) {
		Reqmovr(pp, refD(4), refB(0));
		response = pti_get_response(pp, 8);
	}

	for(i = 0; i < MAXRXBUF; i++) {
		int entry = pti_remove(pp, FREE_CHAIN);
		pti_append(pp, RX_CHAIN, entry);
	}

	for(i = 0; i < MAXTXBUF; i++) {
		int entry = pti_remove(pp, FREE_CHAIN);
		pti_append(pp, TX_CHAIN, entry);
	}

	memset(pp->trch[0].odudb, 0, (BPSIZE * 3));
	rxd = pp->rcvch[0].pspb;
	for(i = 0; i < MINRCH0; i++) {
		unsigned char *buf = pti_get_rx_buff(pp);

		rxd->cnt = BPSIZE;
		if(!i)
			rxd->loc_tag = DESC_FIRST;
		else
			rxd->loc_tag = 0;

		rxd->loc_tag |= (((unsigned int)buf) & DESC_ADDRESS);

		rxd++;
		pp->rcvch[0].psp_rd = (struct psp *)(rxd);
		pp->rcvch[0].pspl++;
		pp->rcvch[0].pspn++;
	}

	rxd = pp->rcvch[1].pspb;
	rxd->cnt = BPSIZE;
	rxd->loc_tag = DESC_FIRST |
		(((unsigned int)pti_get_rx_buff(pp)) & DESC_ADDRESS);
	pp->rcvch[1].psp_rd = (struct psp *)(rxd + 1);
	pp->rcvch[1].pspl++;
	pp->rcvch[1].pspn++;

	rxd = pp->rcvch[2].pspb;
	for(i = 0; i < (MAXRXBUF - MINRCH0 - 1); i++) {
		unsigned char *buf = pti_get_rx_buff(pp);

		rxd->cnt = BPSIZE;
		if(!i)
			rxd->loc_tag = DESC_FIRST;
		else
			rxd->loc_tag = 0;
		rxd->loc_tag |= (((unsigned int)buf) & DESC_ADDRESS);
		rxd++;
		pp->rcvch[2].psp_rd = ((struct psp *)(rxd));
		pp->rcvch[2].pspl++;
		pp->rcvch[2].pspn++;
	}
	pti_set_channel(pp);
}

static void pti_init_rings(struct ptifddi *pp, int from_irq)
{
}

static int pti_init(struct ptifddi *pp, int from_irq)
{
	return 0;
}

static void pti_is_not_so_happy(struct ptifddi *pp)
{
}

static inline void pti_tx(struct ptifddi *pp, struct device *dev)
{
}

static inline void pti_rx(struct ptifddi *pp, struct device *dev)
{
}

static inline void pti_check_limits(struct ptifddi *pp)
{
	int i;

	for(i = 0; i < 3; i++) {
		struct inchan *rv = &pp->rcvch[i];

		if(rv->idudlimon > 64) {
			unsigned int cmd = LIMIT;

			cmd |= ((8 - (2 * i)) << 16);
			cmd |= (rv->idudl & 0x1ff);
			pti_do_request(pp, 0, cmd);
			rv->idudlimon = 0;
		}
		if(rv->psplimon > 5) {
			unsigned int cmd = LIMIT;

			cmd |= ((9 - (2 * i)) << 16);
			cmd |= (rv->pspl & 0x1ff);
			pti_do_request(pp, 0, cmd);
			rv->psplimon = 0;
		}
	}
}

static void pti_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct device *dev		= (struct device *) dev_id;
	struct ptifddi *pp		= (struct ptifddi *) dev->priv;
	unsigned short *iaddr		= (unsigned short *) &pp->dpram->istat[0];
	unsigned short iack, istate;

	DTST(("pti_interrupt: "));
	if(!(*iaddr)) {
		DTST(("iaddr is zero, returning\n"));
		return;
	}

	/* Wait for non-busy state. */
	DTST(("(bloop)..."));
	while(pp->dpram->card_busy)
		barrier();

	/* Tell card that we are looking at things. */
	DTST(("done HBUSY "));
	pp->dpram->host_busy = 1;

	/* Latch interrupt status, clear it for the card. */
	DTST(("istat_latch/clear "));
	istate = pp->dpram->istat[0] | (pp->dpram->istat[1] << 8);
	pp->dpram->istat[0] = pp->dpram->istat[1] = 0;

	/* Acknowledge that we've latches the irq status. */
	iack = pti_get_iack(pp);
	DTST(("iack=%04x istate=%04x ", iack, istate));

	/* Ok, we've latches all of the state we need, let the card go. */
	DTST(("HUNBUSY "));
	pp->host_busy = 0;

	if(!istate) {
		DTST(("zero istate, returning\n"));
		return;
	}

	/* XXX This means the ring is up? */
	if(istate & DIRECTEDOFF) {
		DTST(("DIRECTEDOFF "));
		pp->stati.directe++;
		pti_directboff(pp);
		pti_fddi_restart(pp);
		pp->directbonf = 0;
		pp->fatalerr = 0;
	}

	pp->int_value |= (istate & ~(RXINT | MACRING | MSGARRIV | DIRECTEDOFF));

	if(istate & (MSGARRIV | DIRECTEDON | FATALERR)) {
		DTST(("smt_stuff "));
#if 0
		if(pp->smt_proc && !pp->diagno && pp->sleeping)
			wake_up(pp->smt_wait_queue);
#endif
	}

	if(istate & MACRING) {
		unsigned short response;

		/* Ring status has changed. */
		DTST(("MACRING "));
		response = pti_answer(RINGSTATE);
		pp->fddi_noring = (~(response)) & 0x80;
		pp->stati.ringch++;
		pp->stati.ringstat = response & 0x80;
	}

	if((!norecit) &&
	   (istate & RXINT) &&
	   !pp->fatalerr &&
	   !(pp->int_value & FATALERR)) {
		pti_rx(pp, 0);
		pti_rx(pp, 1);
		pti_rx(pp, 2);
	}

	if(pp->recenb)
		pti_check_limits(pp);
}

static int pti_open(struct device *dev)
{
	struct ptifddi *pp = (struct ptifddi *) dev->priv;

	return pti_init(pp, intr_count);
}

static int pti_close(struct device *dev)
{
#if 0
	struct ptifddi *pp = (struct ptifddi *) dev->priv;
#endif
	return 0;
}

static int pti_start_xmit(struct sk_buff *skb, struct device *dev)
{
#if 0
	struct ptifddi *pp = (struct ptifddi *) dev->priv;
#endif
	return 0;
}

static struct enet_statistics *pti_get_stats(struct device *dev)
{ return &(((struct ptifddi *)dev->priv)->enet_stats); }

static void pti_set_multicast(struct device *dev)
{
}

static int pti_fddi_init(struct device *dev, struct linux_sbus_device *sdev, int num)
{
	static unsigned version_printed = 0;
	struct ptifddi *pp;
	int i, clicks;

	dev = init_etherdev(0, sizeof(struct ptifddi));
	pp = (struct ptifddi *) dev->priv;

	if(version_printed++ == 0)
		printk(version);

	if(sdev->num_registers != 3) {
		printk("ptifddi: num_registers not 3, it is %d\n", sdev->num_registers);
		return ENODEV;
	}

	prom_apply_sbus_ranges(sdev->my_bus, &sdev->reg_addrs[0], sdev->num_registers, sdev);

	/* Register 0 mapping contains DPRAM. */
	pp->dpram = sparc_alloc_io(sdev->reg_addrs[0].phys_addr, 0,
				   sdev->reg_addrs[0].reg_size,
				   "PTI FDDI DPRAM",
				   sdev->reg_addrs[0].which_io, 0);
	if(!pp->dpram) {
		printk("ptiFDDI: Cannot map DPRAM I/O area.\n");
		return ENODEV;
	}

	/* Next, register 1 contains reset byte. */
	pp->reset = sparc_alloc_io(sdev->reg_addrs[1].phys_addr, 0,
				   sdev->reg_addrs[1].reg_size,
				   "PTI FDDI RESET Byte",
				   sdev->reg_addrs[1].which_io, 0);
	if(!pp->reset) {
		printk("ptiFDDI: Cannot map RESET byte.\n");
		return ENODEV;
	}

	/* Register 2 contains unreset byte. */
	pp->unreset = sparc_alloc_io(sdev->reg_addrs[2].phys_addr, 0,
				     sdev->reg_addrs[2].reg_size,
				     "PTI FDDI UNRESET Byte",
				     sdev->reg_addrs[2].which_io, 0);
	if(!pp->unreset) {
		printk("ptiFDDI: Cannot map UNRESET byte.\n");
		return ENODEV;
	}

	/* Acquire supported DMA bursts on this Sbus. */
	pp->bursts = prom_getintdefault(pp->sdev->my_bus->prom_node,
					"burst-sizes", 0x00);

	/* Reset the card. */
	pti_reset(pp);

	request_irq(sdev->irqs[0].pri, &pti_interrupt, SA_SHIRQ,
		    "PTI FDDI", (void *) dev);

	for(i = 0; i < 6; i++)
		dev->dev_addr[i] = idprom->id_ethaddr[i];

	/* Run boot-up card tests. */
	DET(("card_test["));
	i = pti_card_test(pp);
	if(i) {
		DET(("FAILED]\n"));
		printk("ptiFDDI: Bootup card test fails.\n");
		return ENODEV;
	}
	DET(("SUCCESS] "));

	/* Clear DPRAM, start afresh. */
	pti_clear_dpram(pp);

	/* Init the firmware loader. */
	DET(("init_fwloader "));
	pti_init_firmware_loader(pp);

	/* Now load main card FDDI firmware, using the loader. */
	DET(("load_fware "));
	pti_load_main_firmware(pp);

	/* Reset DPRAM logic. */
	DET(("reset_dpram "));
	pp->dpram->reset = DPRAM_RESET_SINGLE;

	if(pti_allocate(pp)) {
		DET(("allocate fails\n"));
		printk("ptifddi: Cannot allocate DMA buffers.\n");
		return ENODEV;
	}

	/* Mark card/host as not busy. */
	DET(("host/card-unbusy "));
	pp->dpram->cbusy = 0;
	pp->dpram->hbusy = 0;

	/* Clear all interrupt status. */
	DET(("irq_clear "));
	pp->dpram->istat = 0;
	pti_get_iack(pp);

	/* Resynchronize. */
	DET(("resync "));
	pp->dpram->re_sync = 0;

	/* Get the board going. */
	DET(("disab/unreset "));
	pp->dpram->disable = 1;
	pti_unreset(pp);

	/* Wait for the card. */
	DET(("waiting... "));
	clicks = 100000;
	while(--clicks) {
		if(pp->dpram->re_sync)
			break;
		udelay(10);
	}

	if(!clicks) {
		DET(("card bolixed\n"));
		printk("ptifddi: Board would not start up.\n");
		return ENODEV;
	}
	DET(("card wokeup, "));

#ifdef DEBUG_DETECT
	dump_dpram(pp);
#endif

	pp->dual  = pp->dpram->dswitch;
	pp->tpair = pp->dpram->tswitch;

	DET(("dual[%02x] tpair[%02x] ", pp->dual, pp->tpair));

	/* Get the firmware IRQ handlers ready to go. */
	DET(("set fw ihandlers, "));
	pti_set_fware_irq_handlers(pp);

	/* Turn on the FDDI state machines etc. */
	DET(("netstart "));
	pti_fddi_go(pp);

	DET(("success\n"));
	return 0;
}

int ptifddi_sbus_probe(struct device *dev)
{
	struct linux_sbus *bus;
	struct linux_sbus_device *sdev = 0;
	static int called = 0;
	int cards = 0, v;

	if(called)
		return ENODEV;
	called++;

	for_each_sbus(bus) {
		for_each_sbusdev(sdev, bus) {
			if(cards) dev = NULL;
			if(!strcmp(sdev->prom_name, "PTI,sbs600") ||
			   !strcmp(sdev->prom_name, "DPV,fddid")) {
				cards++;
				DET(("Found PTI FDDI as %s\n", sdev->prom_name));
				if((v = pti_fddi_init(dev, sdev, (cards - 1))))
					return v;
			}
		}
	}
	if(!cards)
		return ENODEV;
	return 0;
}


#ifdef MODULE

int
init_module(void)
{
	root_pti_dev = NULL;
	return ptifddi_sbus_probe(NULL);
}

void
cleanup_module(void)
{
	struct ptifddi *pp;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	while (root_pti_dev) {
		pp = root_pti_dev->next_module;

		unregister_netdev(root_pti_dev->dev);
		kfree(root_pti_dev->dev);
		root_pti_dev = pp;
	}
}

#endif /* MODULE */
