/* $Id: ptifddi.h,v 1.1.2.4 1997/03/03 07:00:52 davem Exp $
 * ptifddi.c: Defines for Performance Technologies FDDI sbus cards.
 *
 * Copyright (C) 1996 David S. Miller (davem@caip.rutgers.edu)
 */

#ifndef _PTIFDDI_H
#define _PTIFDDI_H

struct dpram_loader {
	volatile unsigned char dpram_stat;		/* 0x000 */
	volatile unsigned char smt_nbytes;		/* 0x001 */
	volatile unsigned char addr_low;		/* 0x002 */
	volatile unsigned char addr_hi;			/* 0x003 */
	volatile unsigned char num_bytes;		/* 0x004 */
	volatile unsigned char data[0x3b];		/* 0x005 */

	volatile unsigned char loader_firmware[0xc0];	/* 0x040 */
};

struct dfddi_ram {
/*0x000*/	volatile unsigned char		regs[0x100];
/*0x100*/	struct dpram_loader		loader;
/*0x200*/	volatile unsigned char		instructions[0x400];
/*0x600*/	volatile unsigned char		msg_in[0x20];
/*0x620*/	volatile unsigned char		msg_out[0x20];
/*0x640*/unsigned char		_unused2[0x50];
/*0x690*/	volatile unsigned char		smsg_in[0x20];
/*0x6b0*/unsigned char		_unused3[0x30];
/*0x6e0*/	volatile unsigned char		beacon_frame[0x20];
/*0x700*/	volatile unsigned char		re_sync;
/*0x701*/unsigned char		_unused4;
/*0x702*/	volatile unsigned char		tswitch;
/*0x703*/unsigned char		_unused5;
/*0x704*/	volatile unsigned short		evq_lost;
/*0x706*/	volatile unsigned short		signal_lost;
/*0x708*/	volatile unsigned short		lerror;
/*0x70a*/	volatile unsigned short		rstate;
/*0x70c*/unsigned char		_unused9[0xa];
/*0x716*/	volatile unsigned short		dswitch;
/*0x718*/unsigned char		_unused10[0x38];
/*0x750*/	volatile unsigned char		cbusy;
/*0x751*/	volatile unsigned char		hbusy;
/*0x752*/	volatile unsigned char		istat[2];
/*0x754*/unsigned char		_unused11[0x2];
/*0x756*/	volatile unsigned short		disable;
/*0x758*/unsigned char		_unused12[0x36];
/*0x78e*/	volatile unsigned char		ucvalid;
/*0x78f*/unsigned char		_unused13;
/*0x790*/	volatile unsigned int		u0addr;
/*0x794*/unsigned char		_unused14[0x10];
/*0x7a4*/	volatile unsigned short		timerint[2];
/*0x7a8*/	volatile unsigned int		P_player;
/*0x7ac*/	volatile unsigned int		Q_player;
/*0x7b0*/	volatile unsigned short		maxint[2];
/*0x7b4*/unsigned char		_unused15[0xa];
/*0x7be*/	volatile unsigned short		reset;
/*0x7c0*/unsigned char		_unused16[0x3c];
/*0x7fc*/	volatile unsigned short		iack;
/*0x7fe*/	volatile unsigned short		loader_addr;
};

#define DPRAM_SIZE		0x800

#define DPRAM_STAT_VALID	0x80
#define DPRAM_STAT_EMPTY	0x00

#define DPRAM_RESET_SINGLE	0x0001
#define DPRAM_RESET_DUAL	0x0005

typedef union addr {
	unsigned char  *b;
	unsigned short *w;
	unsigned int   *l;
} addr;

typedef union inst {
	unsigned char  bytes[4];
	unsigned short shorts[2];
	unsigned int   word;
} inst;

#define INST_CALL	0x00040000
#define CODE_ASM	0x00004000
#define HLOADER		0x0000f740

/* Data types used to communicate directly with the PTI firmware. */
typedef unsigned short entry;

union param {
	unsigned char	bb[4];
	unsigned short	ww[2];
	unsigned int	ll;
};

union addr {
	unsigned char	*b;
	unsigned short	*w;
	unsigned int	*l;
};

struct inst {
	unsigned char	waitflag;
	union param	instc;
	unsigned short	instaddr;
	union param	answer;
	unsigned short	ansaddr;
};

struct radrs {
	unsigned short	trstatus;
	unsigned char	recstatus;
	int		recb_len;
};

union rdatab {
	union param	data[2];
	union radrs	rds;
};

struct response {
	union rdatab	rdb;
	unsigned char	valid;
	unsigned short	event;
};

struct adrs {
	union addr	buftrx;
	int		buff_len;
	unsigned char	trfc;
	union addr	imessage;
	union addr	omessage;
	union addr	mapmessage;
	int		omess_len;
	int		recb_len;
	int		maplength;
	union addr	bufrcv;
	struct response	resp;
};

union datab {
	union param	cdata[4];
	struct adrs	adr;
};

struct command {
	unsigned char	valid;
	unsigned short	cmd;
	unsigned char	userid;
	struct response	*rp;
	union datab	db;
};

struct idud {
	unsigned int status;
	unsigned int loc_tag;
};

struct req {
	unsigned int uid_size;
	unsigned int loc_tag;
};

struct odud {
	unsigned int cnt;
	unsigned int loc_tag;
};

struct cnf {
	unsigned int status;
	unsigned int type;
};

/* Descriptor tag bits. */
#define DESC_FIRST	0x80000000
#define DESC_LAST	0x40000000
#define DESC_ONLY	(DESC_FIRST | DESC_LAST)
#define DESC_ADDRESS	0x0fffffff

struct psp {
	unsigned int cnt;
	unsigned int loc_tag;
};

struct outchan {
	struct odud	*odudb;
	struct odud	*odude;
	struct odud	*odud_rd;
	struct odud	*odud_wr;

	struct cnf	*cnfb;
	struct cnf	*cnfe;
	struct cnf	*cnf_wr;
	unsigned short	cnfl;
	unsigned short	cnfn;
	unsigned char	cnflimon;

	struct req	*reqb;
	struct req	*reqe;
	struct req	*req_rd;
	struct req	*req_wr;
	unsigned short	reql;
	unsigned char	reqlimon;
};

struct inchan {
	struct idud	*idudb;
	struct idud	*idude;
	struct idud	*idud_rd;
	unsigned short	idudl;
	unsigned short	idudlimon;

	struct psp	*pspb;
	struct psp	*pspe;
	struct psp	*psp_rd;
	unsigned short	pspl;
	unsigned short	pspn;
	unsigned char	psplimon;
};

struct commque {
	struct command		cmd;
	struct task_struct	*procp;
	struct commque		*next;
};

struct dfddi_regs {
	volatile unsigned short	*mintr;
	volatile unsigned short	*int_addr;
	volatile unsigned short	*dualflag;
	volatile unsigned short	*optaddr;
	volatile unsigned char	*int_flags;
	addr			message_out;
	addr			message_in;
	addr			symessage_in;
};

#define AREASIZE	100
#define POOLSPACE	(AREASIZE - 14)
#define MAXMAPBUFFER	100
#define BPSIZE		0x1000

#define MAXTXBUF	40
#define MAXRXBUF	(POOLSPACE - MAXTXBUF)

#define MAXUS		5
#define NUMCHAINS	(MAXUS + 3)
#define FREE_CHAIN      0
#define RX_CHAIN        1
#define TX_CHAIN        2

#define MAX_COMMQUEUE	14
#define MINRCH0		30

struct ptifddi {
	struct dfddi_ram	*dpram;
	volatile unsigned char	*reset;
	volatile unsigned char	*unreset;
	volatile unsigned int	*mailbox;
	struct device		*dev;
	unsigned char		tpair, dual, bursts;
	struct enet_statistics	enet_stats;	/* Interface stats.           */
	unsigned int		*dma_block;
	struct ptifddi		*next_module;

	/* XXX Crap which will go in final version. */
	unsigned char		*res;
	unsigned short		*mintr;
	unsigned char		*unres;
	unsigned short		*dpr;
	unsigned short		*dualflag;
	unsigned short		*optaddr;
	union addr		mlbx;
	union addr		dma_base;
	unsigned short		*int_addr;
	unsigned char		*int_flags;
	unsigned short		int_value;
	union addr		message_out;
	union addr		message_in;
	union addr		symessage_in;
	union addr		minstaddr;
	union addr		pcaddr;
	union addr		maddr;
	union inst		rq;
	int			requestbusy;
	int			commexecute;
	struct command		us_comm;
	struct task_struct	*us_proc;
	struct commque		*cdqwr;
	struct commque		*cdqrd;
	int			commquecnt;
	struct commque		*command_queue;
	unsigned char		commandbusy;
	unsigned char		own_address[6];
	unsigned char		*areapoi;
	unsigned char		*area;
	unsigned char		*poolstart;
	unsigned int		buffsize;
	union addr		mailbox;
	struct outchan		trch[2];
	struct inchan		rcvch[3];
	unsigned char		ruid;
	entry			size[NUMCHAINS+1];
	entry			last[NUMCHAINS+1];
	entry			first[NUMCHAINS+1];
	entry			dirtry[POOLSPACE+1];
	unsigned char		ident_count;
	unsigned char		fddi_noring;
	unsigned char		fddi_smt;
	unsigned char		diagno;
	unsigned char		sleeping;
	unsigned short		globevent;

	unsigned char		snaphdr[6];
	int			splvalue;
	unsigned char		spldis;

	/* XXX user_fddi stuff... */

	unsigned char		smt_userchain;
	struct task_struct	*smt_proc;
	unsigned char		directbonf;

	/* XXX statistics... */

	int			recenb;
	int			fatalerr;
	unsigned char		optwis;
	unsigned char		mapreqflag;
	unsigned char		maprespflag;
	unsigned char		mapbuffer[MAXMAPBUFFER];

	/* XXX devinfo */
};

#endif /* !(_PTIFDDI_H) */
