/* $Id: sunfb.c,v 1.13.2.1 1996/11/13 06:56:15 davem Exp $
 * sunfb.c: Sun generic frame buffer support.
 *
 * Copyright (C) 1995, 1996 Miguel de Icaza (miguel@nuclecu.unam.mx)
 * Copyright (C) 1995 David S. Miller (davem@caip.rutgers.edu)
 * 
 * Added getcmap ioctl, may, 96
 * Support for multiple fbs, sep, 96
 */

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kd.h>
#include <linux/malloc.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/types.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/bitops.h>
#include <asm/oplib.h>
#include <asm/sbus.h>
#include <asm/fbio.h>
#include <asm/io.h>
#include <asm/pgtsun4c.h>	/* for the sun4c_nocache */

#include "../../char/kbd_kern.h"
#include "../../char/vt_kern.h"
#include "../../char/consolemap.h"
#include "../../char/selection.h"
#include "../../char/console_struct.h"

#include "fb.h"

extern void set_other_palette (int);
extern void sun_clear_fb(int);
extern void set_cursor (int);

#define FB_SETUP(err) \
	int minor = FB_DEV (inode->i_rdev); \
\
	if (minor >= fbinfos || \
	    fbinfo [minor].type.fb_type == FBTYPE_NOTYPE) \
		return -(err);

static int
fb_open (struct inode * inode, struct file * file)
{
	FB_SETUP(EBADF)
	if (fbinfo [minor].open)
		return -EBUSY;
	fbinfo [minor].mmaped = 0;
	fbinfo [minor].open = 1;
	return 0;
}

static int
fb_ioctl (struct inode *inode, struct file *file, uint cmd, unsigned long arg)
{
	fbinfo_t *fb;
	struct fbcmap *cmap;
	int i;
	FB_SETUP(EBADF)
	
	fb = &fbinfo [minor];
	
	switch (cmd){
	case FBIOGTYPE:		/* return frame buffer type */
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (struct fbtype));
		if (i) return i;
		*(struct fbtype *)arg = (fb->type);
		break;
	case FBIOGATTR:{
		struct fbgattr *fba = (struct fbgattr *) arg;
		
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (struct fbgattr));
		if (i) return i;
		fba->real_type = fb->real_type;
		fba->owner = 0;
		fba->fbtype = fb->type;
		fba->sattr.flags = 0;
		fba->sattr.emu_type = fb->type.fb_type;
		fba->sattr.dev_specific [0] = -1;
		fba->emu_types [0] = fb->type.fb_type;
		for (i = 1; i < 4; i++)
			fba->emu_types [i] = fb->emulations [i];
		break;
	}
	case FBIOSATTR:
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbsattr));
		if (i) return i;
		return -EINVAL;
	case FBIOSVIDEO:
		i = verify_area(VERIFY_READ, (void *)arg, sizeof(int));
		if (i) return i;
		
 		if (fb == fbinfo && vt_cons[fg_console]->vc_mode == KD_TEXT)
 			break;
		if (*(int *)arg){
			if (!fb->blanked || !fb->unblank)
				break;
			if (!minor || (fb->open && fb->mmaped))
				(*fb->unblank)(fb);
			fb->blanked = 0;
		} else {
			if (fb->blanked || !fb->blank)
				break;
			(*fb->blank)(fb);
			fb->blanked = 1;
		}
		break;
	case FBIOGVIDEO:
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (int));
		if (i) return i;
		*(int *) arg = fb->blanked;
		break;
	case FBIOGETCMAP: {
		char *rp, *gp, *bp;
		int end, count;

		if (!fb->loadcmap)
			return -EINVAL;
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbcmap));
		if (i) return i;
		cmap = (struct fbcmap *) arg;
		count = cmap->count;
		if ((cmap->index < 0) || (cmap->index > 255))
			return -EINVAL;
		if (cmap->index + count > 256)
			count = 256 - cmap->index;
		i = verify_area (VERIFY_WRITE, rp = cmap->red, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_WRITE, gp = cmap->green, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_WRITE, bp = cmap->blue, cmap->count);
		if (i) return i;
		end = cmap->index + count;
		for (i = cmap->index; i < end; i++){
			*rp++ = fb->color_map CM(i,0);
			*gp++ = fb->color_map CM(i,1);
			*bp++ = fb->color_map CM(i,2);
		}
		(*fb->loadcmap)(fb, cmap->index, count);
                break;			

	}
	case FBIOPUTCMAP: {	/* load color map entries */
		char *rp, *gp, *bp;
		int end, count;
		
		if (!fb->loadcmap)
			return -EINVAL;
		i = verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbcmap));
		if (i) return i;
		cmap = (struct fbcmap *) arg;
		count = cmap->count;
		if ((cmap->index < 0) || (cmap->index > 255))
			return -EINVAL;
		if (cmap->index + count > 256)
			count = 256 - cmap->index;
		i = verify_area (VERIFY_READ, rp = cmap->red, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_READ, gp = cmap->green, cmap->count);
		if (i) return i;
		i = verify_area (VERIFY_READ, bp = cmap->blue, cmap->count);
		if (i) return i;

		end = cmap->index + count;
		for (i = cmap->index; i < end; i++){
			fb->color_map CM(i,0) = *rp++;
			fb->color_map CM(i,1) = *gp++;
			fb->color_map CM(i,2) = *bp++;
		}
		(*fb->loadcmap)(fb, cmap->index, count);
                break;			
	}

	case FBIOGCURMAX:
		if (!fb->setcursor) return -EINVAL;
		i = verify_area (VERIFY_WRITE, (void *) arg, sizeof (struct fbcurpos));
		if (i) return i;
		((struct fbcurpos *) arg)->fbx = fb->cursor.hwsize.fbx;
		((struct fbcurpos *) arg)->fby = fb->cursor.hwsize.fby;
		break;

	case FBIOSCURSOR:
		if (!fb->setcursor) return -EINVAL;
 		if (fb == fbinfo) {
 			if (vt_cons[fg_console]->vc_mode == KD_TEXT)
 				return -EINVAL; /* Don't let graphics programs hide our nice text cursor */
 			sun_hw_cursor_shown = 0; /* Forget state of our text cursor */
		}
		return sun_hw_scursor ((struct fbcursor *) arg, fb);

	case FBIOSCURPOS:
		if (!fb->setcursor) return -EINVAL;
                /* Don't let graphics programs move our nice text cursor */
 		if (fb == fbinfo) {
 			if (vt_cons[fg_console]->vc_mode == KD_TEXT)
 				return -EINVAL; /* Don't let graphics programs move our nice text cursor */
 		}
		i= verify_area (VERIFY_READ, (void *) arg, sizeof (struct fbcurpos));
		if (i) return i;
		fb->cursor.cpos = *(struct fbcurpos *)arg;
		(*fb->setcursor) (fb);
		break;
		
	default:
		if (fb->ioctl){
			i = fb->ioctl (inode, file, cmd, arg, fb);
			if (i == -ENOSYS) {
				printk ("[[FBIO: %8.8x]]\n", cmd);
				return -EINVAL;
			}
			return i;
		}
		printk ("[[FBIO: %8.8x]]\n", cmd);
		return -EINVAL;
	}
	return 0;
}

static int
fb_close (struct inode * inode, struct file *filp)
{
	fbinfo_t *fb;
	struct fbcursor cursor;
	FB_SETUP(EBADF)	

	fb = &fbinfo[minor];
	
	if (!minor)
		vt_cons [fb->vtconsole]->vc_mode = KD_TEXT;

	/* Leaving graphics mode, turn off the cursor */
	if (fb->mmaped)
		sun_clear_fb (minor);
	cursor.set    = FB_CUR_SETCUR;
	cursor.enable = 0;
	
	/* Reset the driver */
	if (fb->reset)
		fb->reset(fb);

	if (fb->open)
		fb->open = 0;
	fb_ioctl (inode, filp, FBIOSCURPOS, (unsigned long) &cursor);
	set_other_palette (minor);
	if (!minor) {
		render_screen ();
		set_cursor (fg_console);
	} else if (fb->blank)
		(*fb->blank)(fb);
	return 0;
}

static int
fb_mmap (struct inode *inode, struct file *file, struct vm_area_struct *vma)
{
	fbinfo_t *fb;
	FB_SETUP(ENXIO)

	fb = &fbinfo [minor];

	if (fb->mmap){
		int v;
		
		v = (*fb->mmap)(inode, file, vma, fb->base, fb);
		if (v) return v;
		if (!fb->mmaped) {
			fb->mmaped = 1;
			if (!minor) {
				fb->vtconsole = fg_console;
				vt_cons [fg_console]->vc_mode = KD_GRAPHICS;
			} else {
				if (fb->unblank && !fb->blanked)
					(*fb->unblank)(fb);
			}
		}
		return 0;
	} else
		return -ENXIO;
}

static struct file_operations graphdev_fops =
{
	NULL,			/* lseek */
	NULL,			/* read */
	NULL,			/* write */
	NULL,			/* readdir */
	NULL,			/* select */
	fb_ioctl,
	fb_mmap,
	fb_open,		/* open */
	(void(*)(struct inode *, struct file *))
	fb_close,		/* close */
};

__initfunc(int fb_init (void))
{
	/* Register the frame buffer device */
	if (register_chrdev (GRAPHDEV_MAJOR, "graphics", &graphdev_fops)){
		printk ("Could not register graphics device\n");
		return -EIO;
	}
	return 0; /* success */
}    
