/* $Id: vga16.c,v 1.1.2.5 1997/03/18 08:00:15 ecd Exp $
 * vga16.c: Tadpole SPARCbook 1 VGA console driver
 *
 * Copyright (C) 1996 Eddie C. Dost  (ecd@skynet.be)
 */

#include <linux/kd.h>
#include <linux/tty.h>
#include <linux/malloc.h>

#include <asm/sbus.h>
#include <asm/io.h>
#include <asm/fbio.h>
#include <asm/pgtable.h>

#include "../../char/vt_kern.h"
#include "../../char/selection.h"
#include "../../char/console_struct.h"
#include "fb.h"
#include "cg_common.h"


extern unsigned char vga_font[];

struct vga16_regs {
	__volatile__ __u8	crtc_bw_index;
	__volatile__ __u8	crtc_bw_data;
	unsigned char		pad1[10];
	__volatile__ __u8	ac_index;
	__volatile__ __u8	ac_data;
	__volatile__ __u8	misc_out_wr;
	__volatile__ __u8	mb_sleep;
	__volatile__ __u8	seq_index;
	__volatile__ __u8	seq_data;
	__volatile__ __u8	dac_mask;
	__volatile__ __u8	dac_addr_rd;
	__volatile__ __u8	dac_addr_wr;
	__volatile__ __u8	dac_data;
	__volatile__ __u8	feature;
	unsigned char		pad2;
	__volatile__ __u8	misc_out_rd;
	unsigned char		pad3;
	__volatile__ __u8	gc_index;
	__volatile__ __u8	gc_data;
	unsigned char		pad4[4];
	__volatile__ __u8	crtc_index;
	__volatile__ __u8	crtc_data;
	unsigned char		pad5[4];
	__volatile__ __u8	status;
};


static int
vga_extension(fbinfo_t *fb, int on)
{
	struct vga16_regs *regs = fb->info.vga16.regs;
	int old;

	regs->gc_index = 0x0a;
	old = regs->gc_data & 1;
	regs->gc_data = on ? 0xec : 0xce;
	return old;
}


static void
vga16_blitc(unsigned short charattr, int xoff, int yoff)
{
	unsigned char attrib = CHARATTR_TO_SUNCOLOR(charattr);
	unsigned char *font = &vga_font[(charattr & 0xff) << 4];
	struct vga16_regs *regs;
	unsigned char *dst;
	unsigned char fg, bg;
	int i;

	fg = attrib >> 4;
	bg = attrib & 0x0f;

	dst = (unsigned char *)(fbinfo[0].base + 80 * yoff + xoff / 8);
	regs = fbinfo[0].info.vga16.regs;

	regs->gc_index = 0x05;
	regs->gc_data = 0x02;
	regs->gc_index = 0x08;
	for (i = 0; i < 16; i++, font++, dst += 80) {
		__asm__ __volatile__ ("stb	%0, [%1]\n\t"
				      "ldub	[%2], %%g0\n\t"
				      "stb	%3, [%2]\n\t"
				      "stb	%4, [%1]\n\t"
				      "ldub	[%2], %%g0\n\t"
				      "stb	%5, [%2]\n\t" : :
				      "r" (~(*font)), "r" (&regs->gc_data),
				      "r" (dst), "r" (fg),
				      "r" (*font), "r" (bg));
	}
	regs->gc_data = 0xff;
	regs->gc_index = 0x05;
	regs->gc_data = 0x00;
}


static void
vga16_setw(int xoff, int yoff, unsigned short c, int count)
{
	while (count) {
		count--;
		vga16_blitc(c, xoff, yoff);
		xoff += 8;
	}
}

static void
vga16_cpyw(int xoff, int yoff, unsigned short *p, int count)
{
	while (count) {
		count--;
		vga16_blitc(*p++, xoff, yoff);
		xoff += 8;
	}
}

static void
vga16_fill(int attrib, int count, int *boxes)
{
	struct vga16_regs *regs;
	unsigned char *dst;
	int i, w, h;

	regs = fbinfo[0].info.vga16.regs;

	regs->gc_index = 0x00;
	regs->gc_data = attrib & 0x0f;
	regs->gc_index = 0x01;
	regs->gc_data = 0x0f;

	while (count-- > 0) {
		dst = (unsigned char *)(fbinfo[0].base +
					80 * boxes[1] + boxes[0] / 8);
		w = boxes[2] - boxes[0];
		h = boxes[3] - boxes[1];
		for (i = 0; i < h; i++, dst += 80)
			memset(dst, 0, w);
		boxes += 4;
	}

	regs->gc_data = 0x00;
}

static void
vga16_setcursormap(fbinfo_t *fb, unsigned char *red,
		   unsigned char *green, unsigned char *blue)
{
	/* nop */
}

static void
vga16_setcursor(fbinfo_t *fb)
{
	struct cg_cursor *c = &fb->cursor;
	struct vga16_regs *regs = fb->info.vga16.regs;
	unsigned long flags;
	unsigned char *dst;
	unsigned short *src;
	static int cx = -1;
	static int cy = -1;

	save_and_cli(flags);
	if (c->enable) {
		if (cx != -1) {
			src = (unsigned short *)(video_mem_base
					+ (cy >> 4) * video_size_row
					+ (cx >> 2));
			vga16_blitc(*src, cx, cy);
		}

		cx = c->cpos.fbx;
		cy = c->cpos.fby;
		dst = (unsigned char *)(fb->base + 80 * (cy + 14) + cx / 8);

		regs->gc_index = 0x00;
		regs->gc_data = 0x00;	/* Color */
		regs->gc_index = 0x01;
		regs->gc_data = 0x0f;
		__asm__ __volatile__ ("ldub	[%0], %%g0\n\t"
				      "stb	%%g0, [%0]\n\t" : :
				      "r" (dst));
		dst += 80;
		__asm__ __volatile__ ("ldub	[%0], %%g0\n\t"
				      "stb	%%g0, [%0]\n\t" : :
				      "r" (dst));
		regs->gc_data = 0x00;
	} else {
		src = (unsigned short *)(video_mem_base
					+ (c->cpos.fby >> 4) * video_size_row
					+ (c->cpos.fbx >> 2));
		vga16_blitc(*src, c->cpos.fbx, c->cpos.fby);
		cx = cy = -1;
	}
	restore_flags(flags);
}

static void
vga16_setcurshape(fbinfo_t *fb)
{
	/* nop */
}

static void
vga16_loadcmap(fbinfo_t *fb, int index, int count)
{
	struct vga16_regs *regs = fb->info.vga16.regs;
	int i;

	regs->dac_mask = 0xff;
	regs->dac_addr_wr = index;
	for (i = index; count--; i++) {
		regs->dac_data = fb->color_map CM(i,0) >> 2;
		regs->dac_data = fb->color_map CM(i,1) >> 2;
		regs->dac_data = fb->color_map CM(i,2) >> 2;
	}
}

static int
vga16_mmap (struct inode *inode, struct file *file, struct vm_area_struct *vma,
	    long base, fbinfo_t *fb)
{
	uint size, map_offset, r;
	int map_size;
	
	return -ENXIO;

	map_size = size = vma->vm_end - vma->vm_start;
	
	if (vma->vm_offset & ~PAGE_MASK)
		return -ENXIO;

	/* To stop the swapper from even considering these pages */
	vma->vm_flags |= FB_MMAP_VM_FLAGS;

	/* This routine should also map the register if asked for,
	 * but we don't do that yet.
	 */
	map_offset = get_phys ((uint) fb->base);
	r = io_remap_page_range (vma->vm_start, map_offset, map_size,
				 vma->vm_page_prot, fb->space);
	if (r)
		return -EAGAIN;

	vma->vm_inode = inode;
	inode->i_count++;
	return 0;
}

static void
vga16_blank (fbinfo_t *fb)
{
	struct vga16_regs *regs = fb->info.vga16.regs;
	int on;

	on = vga_extension(fb, 1);
	regs->gc_index = 0xa8;
	regs->gc_data = 0x00;
	vga_extension(fb, on);
}

static void
vga16_unblank (fbinfo_t *fb)
{
	struct vga16_regs *regs = fb->info.vga16.regs;
	int on;

	on = vga_extension(fb, 1);
	regs->gc_index = 0xa8;
	regs->gc_data = 0x03;
	vga_extension(fb, on);
}


static void
vga16_show_logo(fbinfo_t *fb, unsigned char *data, int xoff, int yoff)
{
	unsigned char *p = data;
	struct vga16_regs *regs = fb->info.vga16.regs;
	unsigned char *dst;
	int i, j;

	dst = (unsigned char *)(fbinfo[0].base + 80 * yoff + xoff / 8);

	regs->gc_index = 0x05;
	regs->gc_data = 0x02;
	regs->gc_index = 0x08;
	for (i = 0; i < 80; i++) {
		for (j = 0; j < 10; j++) {
			__asm__ __volatile__ ("stb	%0, [%1]\n\t"
					      "ldub	[%2], %%g0\n\t"
					      "stb	%3, [%2]\n\t"
					      "stb	%4, [%1]\n\t"
					      "ldub	[%2], %%g0\n\t"
					      "stb	%5, [%2]\n\t" : :
					      "r" (*p), "r" (&regs->gc_data),
					      "r" (dst), "r" (0x01),
					      "r" (~(*p)), "r" (0x00));
			dst++;
			p++;
		}
		dst += 70;
	}
	regs->gc_data = 0xff;
	regs->gc_index = 0x05;
	regs->gc_data = 0x00;
}


__initfunc(void vga16_setup (fbinfo_t *fb, int slot, uint base,
			     int io, struct linux_sbus_device *dev))
{
	printk ("vga%d at %x,%lx\n", slot, dev->reg_addrs[1].which_io,
		(unsigned long)dev->reg_addrs[1].phys_addr);

	fb->type.fb_cmsize = 256;
	fb->mmap = vga16_mmap;
	fb->loadcmap = vga16_loadcmap;
	fb->reset = 0;
	fb->blank = vga16_blank;
	fb->unblank = vga16_unblank;
	fb->setcursor = vga16_setcursor;
	fb->setcursormap = vga16_setcursormap;
	fb->setcurshape = vga16_setcurshape;
	fb->postsetup = sun_cg_postsetup;
#if 0
	fb->ioctl = 0;
#endif
	fb->blitc = vga16_blitc;
	fb->setw = vga16_setw;
	fb->cpyw = vga16_cpyw;
	fb->fill = vga16_fill;
	fb->show_logo_bw = vga16_show_logo;

	fb->info.vga16.regs = sparc_alloc_io ((void *) base,
		0, sizeof (struct vga16_regs), "vga", io, 0);

	fb->base = (unsigned long) sparc_alloc_io(
			(void *)dev->reg_addrs[1].phys_addr, 0,
			((fb->type.fb_depth * fb->type.fb_height
				* fb->type.fb_width) / 8),
			"vga FB", dev->reg_addrs[1].which_io, 0);

	if (slot && sun_prom_console_id != slot)
		vga16_blank (fb);
}
